from esxi_utils.util.response import Response, ResponsePromise
from esxi_utils.util import exceptions, decorators
from contextlib import contextmanager
from esxi_utils.util import log
import urllib.request
import threading
import pyVmomi
import random
import typing
import time
import ssl
import re

if typing.TYPE_CHECKING:
	from esxi_utils.vm.virtualmachine import VirtualMachine

class GuestTools:
	"""
	Wrapper class for functionality related to VMware Tools on a Virtual Machine. The virtual machine must have VMware tools installed for
	any of the contained functions to work.

	:param vm: A `VirtualMachine` object to wrap. VMware tools must be installed on this virtual machine.
	"""
	def __init__(self, vm: 'VirtualMachine'):
		self._vm = vm
		self._vim_vm = self._vm._vim_vm

	@property
	def running(self) -> bool:
		"""
		Whether or not this Virtual Machine has guest tools running.
		"""
		return self._vim_vm.guest.toolsRunningStatus == "guestToolsRunning"

	def wait(self, retries: int = 120, delay: int = 2) -> bool:
		"""
		Wait for the guest tools to be in the running state.

		:param retries: How many times to retry before exiting.
		:param delay: How long to pause between retries (in seconds).

		:return: A boolean whether or not the state was reached in the allotted amount of retries.
		"""
		for attempt in range(1, retries+1):
			if self.running:
				return True
			log.debug(f"{str(self)} {attempt} failed attempts to connect to guesttools")
			time.sleep(delay)
		return False

	def _assert_available(self):
		"""
		Assert that guest tools is in the running state. This will wait for a small period of time for guest tools to be available,
		and will raise an exception if guest tools does not become available in this timeframe.
		"""
		if not self.wait(retries=30, delay=1):
			raise exceptions.GuestToolsError(self._vm, "VMWare tools is not available or not running")

	def execute_program_async(
		self, 
		username: str, 
		password: str, 
		command: str, 
		cwd: typing.Optional[str] = None, 
		make_output_available: bool = True,
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	)  -> 'GuestToolsResponsePromise':
		"""
		Runs a program in the guest operating system asynchronously.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param command: The command to run. This command should use the full path to the program, e.g. `/usr/bin/date` rather than `date`
		:param cwd: The working directory that the command should be run in.
		:param make_output_available: Whether or not to attempt to make output available by running the command with stdout/stderr redirects. This may not work for all operating
		  systems and is therefore left as an optional feature. 
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 

		:return: A `GuestToolsResponsePromise` object.
		"""
		return GuestToolsResponsePromise(
			tools=self, 
			username=username, 
			password=password, 
			command=command, 
			cwd=cwd, 
			make_output_available=make_output_available,
			out_stream_callback=out_stream_callback
		)

	def execute_program(
		self, 
		username: str, 
		password: str, 
		command: str, 
		cwd: typing.Optional[str] = None,
		timeout: typing.Optional[int] = 120,
		assert_status: typing.Optional[int] = None,
		make_output_available: bool = True,
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	) -> 'Response':
		"""
		Runs a program in the guest operating system. Unlike ``execute_program_async``, this will block until the associated command has finished.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param command: The command to run. This command should use the full path to the program, e.g. `/usr/bin/date` rather than `date`
		:param cwd: The working directory that the command should be run in.
		:param timeout: The command timeout in seconds. Set to 0 or `None` to disable timeout.
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.
		:param make_output_available: Whether or not to attempt to make output available by running the command with stdout/stderr redirects. This may not work for all operating
		  systems and is therefore left as an optional feature. 
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 

		:return: A `Response` object
		"""
		return self.execute_program_async(
			username=username, 
			password=password, 
			command=command, 
			cwd=cwd, 
			make_output_available=make_output_available,
			out_stream_callback=out_stream_callback
		).wait(assert_status=assert_status, timeout=timeout)

	@decorators.retry_on_error([pyVmomi.vim.fault.InvalidState], pause_between_attempts=(0.5, 5.0))
	def list_files(self, username: str, password: str, filepath: str) -> typing.List[str]:
		"""
		List files and directories in the given filepath in the guest operating system. 

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param filepath: The filepath to list.

		:return: A list of files and directories in `filepath`.
		"""
		self._assert_available()

		def get_files_object(path, index):
			return self._service_content.guestOperationsManager.fileManager.ListFilesInGuest(
				vm=self._vim_vm,
				auth=self._get_auth(username=username, password=password),
				filePath=path,
				index=index,
			)

		# First resolve symlinks
		files_object = None
		while True:
			files_object = get_files_object(filepath, 0)
			if len(files_object.files) == 1 and files_object.files[0].type == "symlink":
				filepath = files_object.files[0].attributes.symlinkTarget
			else:
				break

		# Begin reading the files
		# We need to paginate since large responses will result in an error
		files = [ f.path for f in files_object.files ]
		while files_object.remaining != 0:
			files_object = get_files_object(filepath, len(files))
			files.extend([ f.path for f in files_object.files ])

		return [ file for file in files if file not in [".", ".."]]

	@decorators.retry_on_error([pyVmomi.vim.fault.InvalidState], pause_between_attempts=(0.5, 5.0))
	def get_file(self, username: str, password: str, filepath: str, encoding: typing.Optional[str] = "utf-8") -> typing.Union[str, bytes]:
		"""
		Get the content of a file at the given filepath in the guest operating system.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param filepath: The filepath of the file to get.
		:param encoding: The encoding to use when reading the file. If `None`, the raw bytes will instead be returned.

		:return: The contents of the file as a string (or bytes if `encoding=None`)
		"""
		self._assert_available()
		info = self._service_content.guestOperationsManager.fileManager.InitiateFileTransferFromGuest(
			vm=self._vim_vm,
			auth=self._get_auth(username=username, password=password),
			guestFilePath=filepath,
		)

		ctx = ssl.create_default_context()
		ctx.check_hostname = False
		ctx.verify_mode = ssl.CERT_NONE

		url = info.url.replace("*", self._vm._client.hostname)
		try:
			with urllib.request.urlopen(url, context=ctx) as response:
				response_content = response.read()
				if not response.status == 200:
					raise exceptions.GuestToolsError(self._vm, f"Unable to get file {filepath} (status {response.status}): {response_content}")
				if encoding:
					return response_content.decode(encoding=encoding)
				return response_content
		except (urllib.error.URLError, urllib.error.HTTPError) as error:
			raise exceptions.GuestToolsError(self._vm, f"Unable to get file {filepath}: {error.reason}")

	@decorators.retry_on_error([pyVmomi.vim.fault.InvalidState], pause_between_attempts=(0.5, 5.0))
	def write_file(self, username: str, password: str, filepath: str, data: typing.Union[str, bytes]):
		"""
		Write data to a file at the given filepath in the guest operating system.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param filepath: The filepath of the file to write to. The file will be created if it does not exist, or will be overwritten if it does exist.
		:param data: The data to write to the file specified by `filepath`
		"""
		self._assert_available()
		if isinstance(data, str):
			data = str.encode(data)

		url = self._service_content.guestOperationsManager.fileManager.InitiateFileTransferToGuest(
			vm=self._vim_vm,
			auth=self._get_auth(username=username, password=password),
			guestFilePath=filepath,
			fileAttributes=pyVmomi.vim.vm.guest.FileManager.FileAttributes(),
			fileSize=len(data),
			overwrite=True
		)
		url = url.replace("*", self._vm._client.hostname)

		ctx = ssl.create_default_context()
		ctx.check_hostname = False
		ctx.verify_mode = ssl.CERT_NONE

		try:
			# Note: Guest Tools does not appear to allow streaming for files
			# Therefore, the input to `data` for urlopen must be a byte array and not a file-like object
			with urllib.request.urlopen(urllib.request.Request(url, data=data, method='PUT'), context=ctx) as response:
				response_content = response.read()
				if not response.status == 200:
					raise exceptions.GuestToolsError(self._vm, f"Unable to write file {filepath} (status {response.status}): {response_content}")
		except (urllib.error.URLError, urllib.error.HTTPError) as error:
			raise exceptions.GuestToolsError(self._vm, f"Unable to write file {filepath}: {error.reason}")

	@decorators.retry_on_error([pyVmomi.vim.fault.InvalidState], pause_between_attempts=(0.5, 5.0))
	def delete_file(self, username: str, password: str, filepath: str):
		"""
		Delete a file in the guest operating system.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param filepath: The filepath of the file to delete.
		"""
		self._assert_available()
		self._service_content.guestOperationsManager.fileManager.DeleteFileInGuest(
			vm=self._vim_vm,
			auth=self._get_auth(username=username, password=password),
			filePath=filepath
		)

	@decorators.retry_on_error([pyVmomi.vim.fault.InvalidState], pause_between_attempts=(0.5, 5.0))
	def delete_directory(self, username: str, password: str, dirpath: str, recursive: bool = False):
		"""
		Delete a directory in the guest operating system.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param dirpath: The path of the directory to delete.
		:param recursive: If true, all subdirectories are also deleted. If false, the directory must be empty for the operation to succeed. 
		"""
		self._assert_available()
		self._service_content.guestOperationsManager.fileManager.DeleteDirectoryInGuest(
			vm=self._vim_vm,
			auth=self._get_auth(username=username, password=password),
			directoryPath=dirpath,
			recursive=recursive
		)

	@decorators.retry_on_error([pyVmomi.vim.fault.InvalidState], pause_between_attempts=(0.5, 5.0))
	def create_temporary_file(self, username: str, password: str, extension: str = "") -> str:
		"""
		Creates a new unique temporary file for the user to use as needed. The user is responsible for removing it when it is no longer needed. 
		A guest-specific location will be used.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param extension: The extension to give to the new temporary file. 

		:return: The absolute path of the temporary file that is created.
		"""
		self._assert_available()
		extension = f".{extension}" if not extension.startswith(".") else extension
		return self._service_content.guestOperationsManager.fileManager.CreateTemporaryFileInGuest(
			vm=self._vim_vm,
			auth=self._get_auth(username=username, password=password),
			prefix="",
			suffix=extension
		)

	@decorators.retry_on_error([pyVmomi.vim.fault.InvalidState], pause_between_attempts=(0.5, 5.0))
	def create_temporary_directory(self, username: str, password: str, prefix: str = "", suffix: str = "") -> str:
		"""
		Creates a new unique temporary directory for the user to use as needed. The user is responsible for removing it when it is no longer needed. 
		A guest-specific location will be used.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param prefix: The prefix to give to the new temporary directory. 
		:param suffix: The suffix to give to the new temporary directory. 

		:return: The absolute path of the temporary directory that is created.
		"""
		self._assert_available()
		return self._service_content.guestOperationsManager.fileManager.CreateTemporaryDirectoryInGuest(
			vm=self._vim_vm,
			auth=self._get_auth(username=username, password=password),
			prefix=prefix,
			suffix=suffix
		)

	@contextmanager
	def use_temporary_directory(self, username: str, password: str, prefix: str = "", suffix: str = "") -> typing.Generator[str, None, None]:
		"""
		Creates a new unique temporary directory for the user to use as needed. The directory is created as a context manager (`with` statement)
		and will be deleted (along with all contents) when the context is completed. A guest-specific location will be used.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		param prefix: The prefix to give to the new temporary directory. 
		:param suffix: The suffix to give to the new temporary directory. 

		:return: The absolute path of the temporary directory that is created.
		"""
		self._assert_available()
		path = self.create_temporary_directory(username=username, password=password, prefix=prefix, suffix=suffix)
		try:
			yield path
		finally:
			success = False
			for _ in range(20):
				try:
					self.delete_directory(username=username, password=password, dirpath=path, recursive=True)
					success = True
					break
				except Exception:
					time.sleep(5)

			if not success:
				self.delete_directory(username=username, password=password, dirpath=path, recursive=True)

	@contextmanager
	def use_temporary_file(self, username: str, password: str, extension: str = "") -> typing.Generator[str, None, None]:
		"""
		Creates a new unique temporary file for the user to use as needed. The file is created as a context manager (`with` statement)
		and will be deleted when the context is completed. A guest-specific location will be used.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param extension: The extension to give to the new temporary file. 

		:return: The absolute path of the temporary file that is created.
		"""
		self._assert_available()
		path = self.create_temporary_file(username=username, password=password, extension=extension)
		try:
			yield path
		finally:
			success = False
			for _ in range(20):
				try:
					self.delete_file(username=username, password=password, filepath=path)
					success = True
					break
				except Exception:
					time.sleep(5)

			if not success:
				self.delete_file(username=username, password=password, filepath=path)

	def reboot(self):
		"""
		Perform an ESXi reboot of the VM (soft reboot).
		"""
		self._assert_available()
		self._vim_vm.RebootGuest()

	def shutdown(self):
		"""
		Perform a clean shutdown of the VM.
		"""
		self._assert_available()
		self._vim_vm.ShutdownGuest()

	@property
	def networks(self) -> typing.List[typing.Dict[str, typing.Any]]:
		"""
		Get information on the networks for this VM, as reported by guest tools.
		"""
		self._assert_available()
		data = []
		for nicinfo in self._vim_vm.guest.net:
			obj = {
				"network": nicinfo.network,
				"ips": [ ip for ip in nicinfo.ipAddress if re.match(r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$", ip) ],
				"mac": nicinfo.macAddress,
				"connected": nicinfo.connected,
			}
			data.append(obj)
		return data
			
	def ip(self, network: str) -> typing.Union[str, None]:
		"""
		Get the IP of this VM on a specific network, as reported by guest tools.

		:param network: The name of the network.

		:return: The IP, if found, or `None` if not found.
		"""
		self._assert_available()
		for network_data in self.networks:
			if network_data["network"] == network and len(network_data["ips"]) > 0:
				return network_data["ips"][0]
		return None

	def _get_auth(self, username: str, password: str):
		"""
		Get the `NamePasswordAuthentication` object for the username/password.

		:param username: The username to use.
		:param password: The password to use.

		:return: The `NamePasswordAuthentication` object.
		"""
		return pyVmomi.vim.vm.guest.NamePasswordAuthentication(username=username, password=password)

	@property
	def _service_content(self):
		"""
		Get the `ServiceContent` object for the current service instance.

		:return: A `ServiceContent` object.
		"""
		return self._vm._client._service_instance.RetrieveContent()

	def __str__(self):
		return f"<{type(self).__name__} for {self._vm.name}>"

	def __repr__(self):
		return str(self)
	

class GuestToolsResponsePromise(ResponsePromise):
	"""
	Encapsulates a Promise for a result of an GuestTools command.

	:param tools: The GuestTools instance from which this promise originates.
	:param username: The username of the virtual machine user.
	:param password: The password of the virtual machine user.
	:param command: The command to run. This command should use the full path to the program, e.g. `/usr/bin/date` rather than `date`
	:param cwd: The working directory that the command should be run in.
	:param make_output_available: Whether or not to attempt to make output available by running the command with stdout/stderr redirects. This may not work for all operating
	  systems and is therefore left as an optional feature. 
	:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 
	"""

	_tools: GuestTools = None
	"""The GuestTools object to which this promise belongs."""

	_username: str = None
	"""The username for the virtual machine."""

	_password: str = None
	"""The password for the virtual machine."""

	_command: str = None
	"""The original command passed to this promise."""

	_cwd: str = None
	"""The working directory for the command to run."""

	_out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	"""The callback to call when stdout becomes available."""

	_stdoutfile: typing.Optional[str] = None
	"""The file to which stdout will be written. If not ``None``, stdout will be continuously read from this file and this file will be deleted when the promise completes."""

	_stderrfile: typing.Optional[str] = None
	"""The file to which stderr will be written. If not ``None``, stderr will be continuously read from this file and this file will be deleted when the promise completes."""

	_pid: int = None
	"""The process ID for the running process."""

	_resp: typing.Optional[Response] = None
	"""The response of the resolved promise."""

	_exception: typing.Optional[Exception] = None
	"""The exception of the resolved promise."""

	_thread_alive: bool = True
	"""Whether the promise thread should currently be running."""

	_thread: threading.Thread = None
	"""The promise thread."""

	# Based on https://github.com/vmware/vsphere-guest-run
	def __init__(
		self, 
		tools: GuestTools, 
		username: str, 
		password: str, 
		command: str, 
		cwd: typing.Optional[str] = None, 
		make_output_available: bool = True, 
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	):
		super().__init__()
		assert isinstance(username, str), "username must be a string"
		assert isinstance(password, str), "password must be a string"
		assert isinstance(command, str), "command must be a string"
		assert isinstance(cwd, str) or cwd is None, "cwd must be a string or None"
		assert isinstance(make_output_available, bool), "make_output_available must be a boolean"
		assert callable(out_stream_callback) or out_stream_callback is None, "out_stream_callback must be a function or None"
		tools._assert_available()

		self._tools = tools
		self._username = username
		self._password = password
		self._command = command
		self._cwd = cwd
		self._out_stream_callback = out_stream_callback

		# Format arguments
		new_command = self._setup(make_output_available=make_output_available)
		tokens = new_command.split()
		program_path = tokens.pop(0)
		arguments = ''
		for token in tokens:
			arguments += f" {token}"

		# Start the process
		program_spec = pyVmomi.vim.vm.guest.ProcessManager.ProgramSpec(programPath=program_path, arguments=arguments, workingDirectory=self._cwd)
		process_manager = self._tools._service_content.guestOperationsManager.processManager
		creds = self._tools._get_auth(username=username, password=password)

		@decorators.retry_on_error([pyVmomi.vim.fault.InvalidState], pause_between_attempts=(0.5, 5.0))
		def start_program() -> int:
			return process_manager.StartProgramInGuest(self._tools._vim_vm, creds, program_spec)

		self._pid = start_program()
		self._thread = threading.Thread(target=self._run)
		self._thread.start()
		log.debug(f"New GuestTools Promise Object: {str(self)}")

	def _setup(self, make_output_available: bool = False) -> str:
		"""
		Perform any setup necessary prior to starting the command.

		:param make_output_available: Whether or not it is requested to make the output of the command available.
		
		:return: The command to run.
		"""
		command = self._command

		# Handle stdout/stderr
		if make_output_available:
			self._stdoutfile = self._tools.create_temporary_file(self._username, self._password, ".out")
			self._stderrfile = self._tools.create_temporary_file(self._username, self._password, ".err")
			command = f"{command} > {self._stdoutfile} 2> {self._stderrfile}"

		return command

	def _run(self):
		"""
		Main thread.
		"""
		creds = self._tools._get_auth(username=self._username, password=self._password)
		process_manager = self._tools._service_content.guestOperationsManager.processManager
		error_tolerance = 10
		last_stdout = ""
		try:
			# Monitor the process to periodically check if it has completed
			while self._thread_alive:
				try:
					processes = process_manager.ListProcessesInGuest(self._tools._vim_vm, creds, [self._pid])
					if len(processes) == 0:
						raise exceptions.GuestToolsError(self._tools._vm, f"Process not found (pid={self._pid})")
				except Exception as e:
					if error_tolerance == 0:
						raise e
					error_tolerance -= 1
					time.sleep(random.uniform(0.5, 5))
					continue

				process = processes[0]

				# Capture output if requested
				if self._out_stream_callback:
					stdout, _ = self._output()
					new_stdout = stdout[len(last_stdout):]
					if len(new_stdout):
						self._out_stream_callback(new_stdout)
					last_stdout = stdout

				# Check if finished
				if process.exitCode is None:
					time.sleep(random.uniform(0.5, 3))
					error_tolerance = 10
					continue

				# Done
				stdout, stderr = self._output()
				self._resp = self._format_response(Response(cmd=self._command, stdout=stdout, stderr=stderr, status=process.exitCode))
				self._thread_alive = False
		except Exception as e:
			self._exception = e
		finally:
			self._thread_alive = False
			self._cleanup()

	def _format_response(self, response: Response) -> Response:
		"""
		Perform post-processing on a result. This will occur when a command finishes executing but before any files are cleaned up.

		:param response: The result of the command.

		:return: The new result.
		"""
		return response

	def wait(self, assert_status: typing.Optional[int] = None, timeout: typing.Optional[int] = 120) -> Response:
		"""
		Block until the associated command exits, returning the result. 
		
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.
		:param timeout: The command timeout in seconds. Set to 0 or `None` to disable timeout.
		"""
		super().wait()
		assert isinstance(assert_status, int) or assert_status is None, "assert_status must be an integer or None"
		assert isinstance(timeout, int) or timeout is None, "timeout must be an integer or None"

		if self._thread_alive:
			self._thread.join(timeout if timeout > 0 else None)
			if self._thread.is_alive():
				# Timed out
				self._terminate()
				self._cleanup()
				raise TimeoutError(f"{str(self)} Guest command \"{self._command}\" timed out after {str(timeout)} seconds")
			
		if self._exception:
			raise self._exception
		
		if assert_status is not None and self._resp and self._resp.status != assert_status:
			raise exceptions.GuestToolsError(self._tools._vm, f"Command failed: {str(self._resp)}")

		return self._resp

	def cancel(self):
		"""
		Cancel the underlying operation.
		"""
		super().cancel()
		self._thread_alive = False
		self._thread.join()
		self._terminate()
		self._cleanup()

	def _terminate(self):
		"""
		Forcefully terminate the underyling process.
		"""
		creds = self._tools._get_auth(username=self._username, password=self._password)
		process_manager = self._tools._service_content.guestOperationsManager.processManager
		@decorators.retry_on_error([pyVmomi.vim.fault.InvalidState], pause_between_attempts=(0.5, 5.0))
		def terminate():
			process_manager.TerminateProcessInGuest(self._tools._vim_vm, creds, self._pid)

		try:
			terminate()
		except pyVmomi.vim.fault.GuestProcessNotFound as e1:
			log.debug(f'Caught vim.fault.GuestProcessNotFound exception from pyVmomi: {str(e1)}')
		except BaseException as e:
			if not str(e).startswith('process not found ('):
				raise exceptions.GuestToolsError(self._tools._vm, f"Failed to terminate process for \"{self._command}\": {str(e)}")

	def _cleanup(self):
		"""
		Clean up any files or other resources after the process has finished.
		"""
		if self._stdoutfile:
			self._tools.delete_file(self._username, self._password, self._stdoutfile)
			self._stdoutfile = None
		if self._stderrfile:
			self._tools.delete_file(self._username, self._password, self._stderrfile)
			self._stderrfile = None

	def _output(self) -> typing.Tuple[str, str]:
		"""
		Get the current stdout/stderr output.

		:return: A tuple (stdout, stderr)
		"""
		stdout = ""
		stderr = ""
		if self._stdoutfile:
			stdout = self._tools.get_file(self._username, self._password, self._stdoutfile)
		if self._stderrfile:
			stderr = self._tools.get_file(self._username, self._password, self._stderrfile)
		return (stdout, stderr)

	def __str__(self):
		return repr(self)

	def __repr__(self):
		return f"<{type(self).__name__} for \"{self._command}\">"


from esxi_utils.util import log, exceptions
from datetime import datetime
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.vm.virtualmachine import VirtualMachine

class SnapshotList:
	"""
	Class representing the snapshots for the given VirtualMachine.

	:param vm: A `VirtualMachine` object
	"""
	def __init__(self, vm: 'VirtualMachine'):
		self._vm = vm
		self._vim_vm = vm._vim_vm

	def __iter__(self) -> typing.List['Snapshot']:
		if self.root is None:
			return iter([])
		return iter(self.root.flattened)

	def __getitem__(self, name_or_id: typing.Union[str, int]) -> 'Snapshot':
		return self.get(name_or_id)

	def find(self, name_or_id: typing.Union[str, int]) -> typing.Union['Snapshot', None]:
		"""
		Find a snapshot by name or ID.
		If multiple snapshots match, this will raise a `MultipleSnapshotsFoundError` exception.

		:param name_or_id: The snapshot name (string) or ID (int).

		:return: A `Snapshot` object for the requested snapshot, or `None` if the snapshot was not found.
		"""
		assert isinstance(name_or_id, (str, int))
		found = [ s for s in self if s.name == name_or_id or s.id == name_or_id ]
		if len(found) == 0:
			return None
		if len(found) > 1:
			raise exceptions.MultipleSnapshotsFoundError(self._vm, name_or_id)
		return found[0]

	def get(self, name_or_id: typing.Union[str, int]) -> 'Snapshot':
		"""
		Find a snapshot by name or ID and raise an exception if not found.
		If multiple snapshots match, this will raise a `MultipleSnapshotsFoundError` exception.

		:param name_or_id: The snapshot name (string) or ID (int).

		:return: A `Snapshot` object for the requested snapshot
		"""
		snapshot = self.find(name_or_id)
		if snapshot is None:
			raise exceptions.SnapshotNotFoundError(self._vm, name_or_id)
		return snapshot

	@property
	def root(self) -> typing.Union['Snapshot', None]:
		"""
		The root snapshot.
		"""
		snapshot_data = self._vim_vm.snapshot
		if snapshot_data is None:
			return None
		return Snapshot(vm=self._vm, id=snapshot_data.rootSnapshotList[0].id)

	@property
	def exists(self) -> bool:
		"""
		Returns whether or not any snapshots exist on this VirtualMachine.

		:return: A boolean whether or not this VirtualMachine has snapshots.
		"""
		return len(list(self)) != 0

	@property
	def current(self) -> typing.Union['Snapshot', None]:
		"""
		Get the current snapshot of this VM.

		:return: A `Snapshot` object for the current snapshot, or `None` if there is no current snapshot.
		"""
		current_snapshot = self._vim_vm.snapshot.currentSnapshot
		current = [ s for s in self if str(s._obj.snapshot) == str(current_snapshot) ]
		if len(current) == 0:
			return None
		return current[0]

	def remove_all(self):
		"""
		Remove all snapshots from this VirtualMachine.
		"""
		if self.root is None:
			return
		self.root.remove(remove_children=True)

	def create(self, name: str, description: str = "", include_memory: bool = False, quiesce: bool = False):
		"""
		Create a new snapshot of this VirtualMachine.

		:param name: 
			The name for this snapshot. The name need not be unique for this virtual machine.
		:param description: 
			A description for this snapshot.
		:param include_memory: 
			If `True`, a dump of the internal state of the virtual machine (basically a memory dump) is included in the snapshot. 
			Memory snapshots consume time and resources, and thus take longer to create.
			When `False`, the power state of the snapshot is set to powered off.
		:param quiesce:
			If `True` and the virtual machine is powered on when the snapshot is taken, VMware Tools is used to quiesce the file system in the virtual machine. 
			This assures that a disk snapshot represents a consistent state of the guest file systems. 
			If the virtual machine is powered off or VMware Tools are not available, the quiesce flag is ignored. 
		"""
		log.info(f"{str(self._vm)} Creating snapshot \"{name}\" (description = \"{description}\", include memory = {str(include_memory)}, quiesce = {str(quiesce)})")
		self._vm._client._wait_for_task(
			self._vim_vm.CreateSnapshot_Task(name=name, description=description, memory=include_memory, quiesce=quiesce)
		)

	def __str__(self):
		return f"<{type(self).__name__} for {self._vm.name} ({len(list(self))} snapshots)>"
	
	def __repr__(self):
		return self.__str__()


class Snapshot:
	"""
	Class representing a single snapshot for a VirtualMachine.
	"""
	def __init__(self, vm: 'VirtualMachine', id: int):
		self._vm = vm
		self._id = id

	@property
	def id(self) -> int:
		"""
		The snapshot ID.
		"""
		return self._id

	@property
	def name(self) -> str:
		"""
		The snapshot name.
		"""
		return self._obj.name

	@property
	def description(self) -> str:
		"""
		The snapshot description.
		"""
		return self._obj.description

	@property
	def createtime(self):
		"""
		The snapshot ID.
		"""
		return datetime.fromisoformat(str(self._obj.createTime))

	@property
	def state(self) -> str:
		"""
		The power state of the virtual machine when this snapshot was taken. 
		"""
		return self._obj.state

	@property
	def powered_on(self) -> bool:
		"""
		Whether the power state of this snapshot is set to on.
		"""
		return self.state == "poweredOn"

	@property
	def quiesced(self) -> bool:
		"""
		Flag to indicate whether or not the snapshot was created with the "quiesce" option, ensuring a consistent state of the file system.
		"""
		return self._obj.quiesced

	@property
	def children(self) -> typing.List['Snapshot']:
		"""
		Child snapshots of this snapshot.
		"""
		return [ Snapshot(vm=self._vm, id=child.id) for child in self._obj.childSnapshotList ]

	@property
	def flattened(self) -> typing.List['Snapshot']:
		"""
		A flat list of snapshot objects containing this snapshot and its children.
		"""
		child_list = list()
		for child in self.children:
			child_list.extend(child.flattened)
		return [self, *child_list]

	def remove(self, remove_children: bool = False):
		"""
		Remove this snapshot (and optionally its children) from the VM.

		:param remove_children: Whether or not to remove children snapshots.
		"""
		self._vm._client._wait_for_task(
			self._obj.snapshot.RemoveSnapshot_Task(removeChildren=remove_children)
		)

	def revert(self, suppress_power_on: bool = False):
		"""
		Revert to this snapshot.

		:param suppress_power_on: If set to true, the virtual machine will not be powered on regardless of the power state when the snapshot was created.
		"""
		self._vm._client._wait_for_task(
			self._obj.snapshot.RevertToSnapshot_Task(suppressPowerOn=suppress_power_on)
		)

	@property
	def _obj(self):
		snapshot_data =  []
		if self._vm._vim_vm.snapshot:
			snapshot_data.extend(self._vm._vim_vm.snapshot.rootSnapshotList)
		for snapshot in snapshot_data:
			if snapshot.id == self.id:
				return snapshot
			snapshot_data.extend(snapshot.childSnapshotList)
		raise exceptions.SnapshotNotFoundError(self._vm, self.id)

	def __str__(self):
		return f"<{type(self).__name__} \"{self.name}\" (ID {self.id}) for {self._vm.name} ({len(self.children)} children)>"
	
	def __repr__(self):
		return self.__str__()

	


from esxi_utils.util import exceptions
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.client import ESXiClient
	from esxi_utils.networking.portgroup import PortGroup

class VMKernelNICList:
	"""
	The list of VMKernel NICs on the ESXi host.
	"""
	def __init__(self, client: 'ESXiClient'):
		self._client = client

	def __iter__(self) -> typing.Iterator['VMKernelNIC']:
		return iter([ VMKernelNIC(self._client, vnic.device) for vnic in self._network_system.networkInfo.vnic ])

	def __getitem__(self, name) -> 'VMKernelNIC':
		return self.get(name)

	def __contains__(self, name: str) -> bool:
		return self.exists(name)

	@property
	def items(self) -> typing.List['VMKernelNIC']:
		"""
		A list of all items
		"""
		return list(self)

	def find(self, name: str) -> typing.Union['VMKernelNIC', None]:
		"""
		Get a VMKernel NIC by name.

		:param name: The name of the VMKernel NIC.

		:return: A `VMKernelNIC` object, or `None` if not found.
		"""
		found = [ vnic for vnic in self if vnic.name == name ]
		if len(found) != 1:
			return None
		return found[0]

	def get(self, name: str) -> 'VMKernelNIC':
		"""
		Get a VMKernel NIC by name and raise an exception if not found.

		:param name: The name of the VMKernel NIC.

		:return: A `VMKernelNIC` object
		"""
		vnic = self.find(name)
		if vnic is None:
			raise exceptions.NetworkingObjectNotFoundError('VMKernelNIC', name)
		return vnic

	def exists(self, name: str) -> bool:
		"""
		Check whether a VMKernel NIC exists 

		:parma name: The name of the VMKernel NIC

		:return: Whether or not the VMKernel NIC exists
		"""
		return self.find(name) is not None

	@property
	def _network_system(self):
		return self._client._host_system.configManager.networkSystem

	def __str__(self):
		return f"<{type(self).__name__} ({len(list(self))} VMKernel NICs)>"
	
	def __repr__(self):
		return str(self)


class VMKernelNIC:
	"""
	A host virtual NIC that provides access to the external network through a virtual switch that is bridged through a VMKernel NIC to a physical network. 
	"""
	def __init__(self, client: 'ESXiClient', name: str):
		self._client = client
		self._name = name

	@property
	def name(self) -> str:
		"""
		The device name of this VMkernel NIC.
		"""
		return self._name

	@property
	def portgroup(self) -> typing.Union['PortGroup', None]:
		"""
		If the VMKernel NIC is connected to a vSwitch, this property is the `PortGroup` connected.
		"""
		from esxi_utils.networking.portgroup import PortGroup
		if self._obj.portgroup and len(self._obj.portgroup):
			return PortGroup(self._client, self._obj.portgroup)
		return None

	@property
	def mac(self) -> typing.Union[str, None]:
		"""
		Media access control (MAC) address of the virtual network adapter. 
		"""
		return self._obj.spec.mac

	@property
	def ip(self) -> typing.Union[str, None]:
		"""
		The IPv4 address currently used by the network adapter.
		"""
		if not self._obj.spec.ip:
			return None
		return self._obj.spec.ip.ipAddress

	@property
	def subnetmask(self) -> typing.Union[str, None]:
		"""
		The subnet mask, specified specified using IPv4 dot notation. 
		"""
		if not self._obj.spec.ip:
			return None
		return self._obj.spec.ip.subnetMask

	@property
	def gateway(self) -> typing.Union[str, None]:
		"""
		The default gateway address.
		"""
		if not self._obj.spec.ipRouteSpec or not self._obj.spec.ipRouteSpec.ipRouteConfig:
			return None
		return self._obj.spec.ipRouteSpec.ipRouteConfig.defaultGateway

	@property
	def mtu(self) -> int:
		"""
		Maximum transmission unit for packets size in bytes for the virtual NIC.
		"""
		return self._obj.spec.mtu

	@property
	def _obj(self):
		for vnic in self._network_system.networkInfo.vnic:
			if vnic.device == self.name:
				return vnic
		raise exceptions.NetworkingObjectNotFoundError('VMKernelNIC', self.name)

	@property
	def _network_system(self):
		return self._client._host_system.configManager.networkSystem

	def __str__(self):
		return f"<{type(self).__name__} \"{self.name}\">"
	
	def __repr__(self):
		return str(self)


from re import L
from lxml import etree
import typing

class XmlFile:
	"""
	Proxy class for working with an XML file.
	"""
	def __init__(self, file, update_callback: typing.Optional[typing.Callable[[str], None]] = None):
		parser = etree.XMLParser(remove_blank_text=True)
		self._tree = etree.parse(file, parser)
		self._update_callback = update_callback

	def update(self):
		if self._update_callback and callable(self._update_callback):
			self._update_callback(self.xml(pretty_print=True, xml_declaration=True))

	def xml(self, pretty_print: bool = True, xml_declaration: bool = False) -> str:
		"""
		Dump the current XML tree to a string representation.

		:param pretty_print: Whether or not to pretty print the XML.
		:param xml_declaration: Whether or not to include the XML declaration.

		:return: The XML string for this node.
		"""
		return etree.tostring(self._tree, encoding="utf-8", pretty_print=pretty_print, xml_declaration=xml_declaration, with_comments=True).decode("utf-8").strip()

	@property
	def root(self):
		return _XmlElement(self._tree.find("."), self)

	def __str__(self):
		return self.xml()

	def __repr__(self):
		return f"<{type(self).__name__} at {self._tree.docinfo.URL}>"
		

class _XmlElement:
	"""
	Proxy class for working with an element of an XML file.
	"""
	def __init__(self, element: 'etree._Element', file: 'XmlFile'):
		assert isinstance(element, etree._Element), "element is not an element"
		self._element = element
		self._file = file

	@property
	def tag(self) -> str:
		namespaces = self._element.nsmap
		prefix = self._element.prefix
		tag = self._element.tag
		if prefix not in namespaces:
			return tag
		uri = namespaces[prefix]
		base = tag[len("{"+uri+"}"):]
		if not prefix:
			return base
		return f"{prefix}:{base}"

	@property
	def children(self) -> typing.List['_XmlElement']:
		return [ _XmlElement(child, self._file) for child in self._element ]

	@property
	def attributes(self) -> '_XmlAttributes':
		return _XmlAttributes(self)

	@property
	def text(self) -> typing.Union[str, None]:
		return self._element.text

	@text.setter
	def text(self, value):
		self._element.text = value
		self._file.update()

	@property
	def summary(self) -> typing.Dict[str, typing.Any]:
		return { "tag": self.tag, "attributes": self.attributes, "text": self.text }
	
	def find(self, tag: str) -> typing.List['_XmlElement']:
		"""
		Find all children matching the given tag name.

		:param tagname: The name of the tag for the child to get.

		:return: A list of `_XmlElement` elements
		"""
		return [ child for child in self.children if child.tag == tag ]

	def get(self, tag: str) -> '_XmlElement':
		"""
		Get the first child matching the given tag name. Raises an exception if the child was not found.

		:param tag: The name of the tag for the child to get.

		:return: An `_XmlElement` element
		"""
		matched = self.find(tag)
		if len(matched) == 0:
			raise KeyError(f"child tag '{tag}' not found")
		return matched[0]

	def append(self, tag: str) -> '_XmlElement':
		"""
		Create a new child element at the end of this element's children.

		:param tag: The name of the new tag to add.

		:return: An `_XmlElement` element for the new child
		"""
		for ns, uri in self._element.nsmap.items():
			if ns is None:
				continue
			prefix = f"{ns}:"
			if tag.startswith(prefix):
				base = tag[len(prefix):]
				tag = "{" + uri + "}" + base
		el = etree.SubElement(self._element, tag)
		self._file.update()
		return _XmlElement(el, self._file)

	def remove(self):
		"""
		Remove this XML node from the tree.
		"""
		self._element.getparent().remove(self._element)
		self._file.update()

	def xml(self, pretty_print: bool = True) -> str:
		"""
		Dump the given element to an XML string representation.

		:param pretty_print: Whether or not to pretty print the XML.

		:return: The XML string for this element.
		"""
		return etree.tostring(self._element, encoding="utf-8", pretty_print=pretty_print, xml_declaration=False, with_comments=True).decode("utf-8").strip()

	def __getitem__(self, tag):
		return self.find(tag)

	def __iter__(self):
		return iter(self.children)

	def __str__(self):
		return self.xml()

	def __repr__(self):
		return str(self.summary)


class _XmlAttributes:
	"""
	Proxy class for working with the attributes on an XML tag
	"""
	def __init__(self, parent: '_XmlElement'):
		self._parent = parent

	def _uri_to_prefix(self, value: str) -> str:
		for ns, uri in self._parent._element.nsmap.items():
			if ns is None:
				continue
			prefix = "{"+uri+"}"
			if value.startswith(prefix):
				base = value[len(prefix):]
				return f"{ns}:{base}"
		return value
	
	def _prefix_to_uri(self, value: str) -> str:
		for ns, uri in self._parent._element.nsmap.items():
			if ns is None:
				continue
			prefix = f"{ns}:"
			if value.startswith(prefix):
				base = value[len(prefix):]
				return "{" + uri + "}" + base
		return value

	@property
	def items(self) -> typing.Dict[str, str]:
		return { self._uri_to_prefix(key): value for key, value in self._parent._element.items() }

	def __getitem__(self, key):
		return self.items[key]

	def __setitem__(self, key, newvalue):
		self._parent._element.set(self._prefix_to_uri(key), newvalue)
		self._parent._file.update()

	def __delitem__(self, key):
		del self._parent._element.attrib[self._prefix_to_uri(key)]
		self._parent._file.update()

	def __str__(self):
		return str(self.items)

	def __repr__(self):
		return str(self)




from esxi_utils.util.connect.unix import UnixSSHConnection
from esxi_utils.util.parse import vimobj
from lxml import etree
import typing
import re


class ESXiSSHConnection(UnixSSHConnection):
	"""
	A SSH connection to a ESXi remote host.

	:param ip: The IP of the target.
	:param username: The username to use when logging into the remote system.
	:param password: The password to use when logging into the remote system.
	:param sudo_password: The password to use when running as sudo. Only necessary if the sudo password differs from `password`, as `password` will be used as the default.
	:param port: The port to use when connecting over SSH.
	"""

	def vimcmd(self, cmd: str, parse: bool = False, include_dtype: bool = True) -> typing.Union[str, typing.Dict[str, typing.Any]]:
		"""
		Executes a `vim-cmd` command on the ESXi server.

		:param cmd: Command to execute for `vim-cmd`
		:param parse: Attempt to parse the output as structured data. Only some commands can be parsed in this way; it is up to the caller to ensure that the command provided is one that returns a parsable format when this is `True`.
		:param include_dtype: Include object types when parse=`True`. For objects with a known and predictable structure, setting this to `False` may improve the usability/readability of the result.

		:return: A string of the command output if parse=`False`. If parse=`True`, a dict will be returned containing the parsed data.
		"""
		assert isinstance(cmd, str)
		cmd = re.sub("^vim-cmd", "", cmd.strip(), 1).strip()
		vimcmd_output = self.exec(f"vim-cmd {cmd}", assert_status=0).stdout
		return vimobj(vimcmd_output, include_dtype=include_dtype) if parse else vimcmd_output


	def esxcli(self, cmd: str, raw: bool = False)-> typing.Union[str, typing.Dict[str, typing.Any], typing.List[typing.Any]]:
		"""
		Executes a `esxcli` command on the remote server. Note: `esxcli` typically gives information about running VMs only.

		:param cmd: Command to execute for `esxcli`.
		:param raw: If True, return the raw (unparsed) output of the command as a string. Otherwise, automatically parse the output.

		:return: A string if `raw=True`, otherwise depends on the command output (common outputs are list or dict).
		"""
		assert isinstance(cmd, str)
		cmd = re.sub(r"^esxcli(\s*--formatter=\S*)?", "", cmd.strip(), 1).strip()
		cmd = f"esxcli {cmd}" if raw else f"esxcli --formatter=xml {cmd}"
		output = self.exec(cmd, assert_status=0).stdout
		if raw:
			return output

		def parse_xml_data(node):
			if node.tag == "list":
				data = list()
				for elem in node.findall("*"):
					data.append(parse_xml_data(elem))
				return data

			if node.tag == "structure":
				data = dict()
				for elem in node.findall("field"):
					field_name = elem.get("name")
					data[field_name] = parse_xml_data(elem)
				return data

			if node.tag == "field":
				field_value = node.find("*[1]")
				return parse_xml_data(field_value)

			if node.tag == "integer":
				return int(node.text)

			text = node.text
			if text is None:
				return ''
			return str(text)

		tree = etree.fromstring(output.encode())

		# Remove namespaces to make the tree easier to parse
		for elem in tree.getiterator(): 
			if not isinstance(elem, etree._Comment) and not isinstance(elem, etree._ProcessingInstruction):
				elem.tag = etree.QName(elem).localname

		# Parse
		root = tree.find(".//root")
		return parse_xml_data(root.find("*[1]"))


from esxi_utils.util import log, exceptions
from lxml import etree
import pyVmomi
import typing
import re

if typing.TYPE_CHECKING:
	from esxi_utils.firewall.firewall import Firewall
	from esxi_utils.firewall.ruleset import Ruleset

class RuleList:
	"""
	List of rules for a ruleset on the host firewall.
	"""
	def __init__(self, ruleset: 'Ruleset'):
		self._ruleset = ruleset

	def __iter__(self) -> typing.Iterator['Rule']:
		return iter([ Rule(self._ruleset, obj) for obj in self._ruleset._obj.rule ])

	def __getitem__(self, index) -> 'Rule':
		return self.items[index]

	@property
	def items(self) -> typing.List['Rule']:
		"""
		A list of all items
		"""
		return list(self)

	def add(self, direction: str, protocol: str, porttype: str, begin: str, end: typing.Optional[str] = None) -> 'Rule':
		"""
		Add a new rule to this ruleset.
		
		:param direction: The new rule's direction (i.e. 'outbound' or 'inbound').
		:param protocol: The new rule's protocol (i.e. 'tcp' or 'udp')
		:param porttype: The new rule's port type (i.e. 'src' or 'dst')
		:param begin: The new rule's begin port.
		:param end: The new rule's end port. If `None`, this rule will only apply to the port denoted by `begin`.

		:return: The newly added `Rule` object
		"""
		assert direction in ["outbound", "inbound"], "direction must be either 'outbound' or 'inbound'"
		assert protocol in ["tcp", "udp"], "protocol must be either 'tcp' or 'udp'"
		assert porttype in ["src", "dst"], "porttype must be either 'src' or 'dst'"
		assert isinstance(begin, int) and begin > 0 and begin <= 65535, "begin must be a integer in range 1-65535"
		assert end is None or (isinstance(end, int) and end > 0 and end <= 65535 and end >= begin), "end must be a integer in range 1-65535 and greater or equal to begin, or None"
		rule_str = f"{direction} {porttype} {str(begin) + '-' + str(end) if end else str(begin)} {protocol}"
		log.info(f"{str(self)} Adding new rule: \"{rule_str}\"")

		def get_matching():
			results = [ rule for rule in self if rule.direction == direction and rule.protocol == protocol and rule.porttype == porttype and rule.port == begin and rule.endport == end ]
			if len(results) == 0:
				return None
			return results[0]

		# Ensure this rule doesn't exist
		if get_matching():
			raise exceptions.FirewallError(self, f"Rule \"{rule_str}\" already exists")

		# Find the ruleset
		tree, existing_ruleset_tag = self._ruleset._get_xml_ruleset()
		existing_rule_elements = existing_ruleset_tag.findall("./rule")
		
		# Re-create the ruleset element
		# To ensure this is properly formatted, as rulesets may be formatted differently (i.e. without rule IDs)
		# we'll just completely re-create this service element by copying over old information alongside the newly added rule
		new_ruleset_tag = etree.Element("service")
		new_ruleset_tag.set("id", existing_ruleset_tag.get("id"))
		etree.SubElement(new_ruleset_tag, "id").text = existing_ruleset_tag.find("./id").text

		new_rule_tag = etree.Element("rule")
		etree.SubElement(new_rule_tag, "direction").text = direction
		etree.SubElement(new_rule_tag, "protocol").text = protocol
		etree.SubElement(new_rule_tag, "porttype").text = porttype

		if end:
			port_tag = etree.SubElement(new_rule_tag, "port")
			etree.SubElement(port_tag, "begin").text = str(begin)
			etree.SubElement(port_tag, "end").text = str(end)
		else:
			etree.SubElement(new_rule_tag, "port").text = str(begin)

		for i, rule in enumerate([*existing_rule_elements, new_rule_tag]):
			rule.set("id", str(i).zfill(4))
			new_ruleset_tag.append(rule)

		etree.SubElement(new_ruleset_tag, "enabled").text = existing_ruleset_tag.find("./enabled").text
		etree.SubElement(new_ruleset_tag, "required").text = existing_ruleset_tag.find("./required").text
		
		# Overwrite the existing ruleset
		tree.replace(existing_ruleset_tag, new_ruleset_tag)
		self._ruleset._firewall._update_service_xml(tree)

		# Get the newly added rule
		matching = get_matching()
		if matching is None:
			raise exceptions.FirewallError(self, f"Failed to get newly added rule \"{rule_str}\"")
		return matching

	def __str__(self):
		return f"<{type(self).__name__} for Ruleset {self._ruleset.key} ({len(self.items)} rules)>"

	def __repr__(self):
		return str(self)


class Rule:
	"""
	A Firewall Rule.
	"""
	def __init__(self, ruleset: 'Ruleset', rule_obj: pyVmomi.vim.host.Ruleset.Rule):
		self._ruleset = ruleset
		self._obj = rule_obj

	@property
	def port(self) -> int:
		"""
		The firewall rule port number or start port.
		"""
		return self._obj.port

	@property
	def endport(self) -> typing.Union[int, None]:
		"""
		For a port range, the ending port number. 
		"""
		return self._obj.endPort

	@property
	def range(self) -> str:
		"""
		The port range as a string if this rule defines a range, otherwise just the port this rule defines.
		"""
		return f"{self.port}-{self.endport}" if self.endport else str(self.port)

	@property
	def direction(self) -> str:
		"""
		The port direction.
		"""
		return self._obj.direction

	@property
	def porttype(self) -> str:
		"""
		The firewall rule port type.
		"""
		return self._obj.portType

	@property
	def protocol(self) -> str:
		"""
		The firewall rule protocol.
		"""
		return self._obj.protocol

	def remove(self):
		"""
		Remove this rule from its parent ruleset.
		"""
		tree, ruleset, rule = self._get_xml_rule()
		ruleset.remove(rule)
		for i, rule in enumerate(ruleset.findall(f"./rule")): # Re-write the IDs to ensure they are correctly ordered
			rule.set("id", str(i).zfill(4))
		self._ruleset._firewall._update_service_xml(tree)

	def _get_xml_rule(self):
		"""
		Get this rule from /etc/vmware/firewall/service.xml as an `etree.Element` object.

		:return: A tuple (tree, ruleset element, rule element)
		"""
		tree, ruleset = self._ruleset._get_xml_ruleset()
		for rule in ruleset.findall(f"./rule"):
			direction = rule.find("./direction").text if rule.find("./direction") is not None else None
			porttype = rule.find("./porttype").text if rule.find("./porttype") is not None else None
			protocol = rule.find("./protocol").text if rule.find("./protocol") is not None else None
			port = rule.find("./port").text if rule.find("./port") is not None else None
			begin = rule.find("./port/begin").text if rule.find("./port/begin") is not None else None
			endport = rule.find("./port/end").text if rule.find("./port/end") is not None else None
			port = begin or port
			if direction != self.direction or porttype != self.porttype or protocol != self.protocol or str(port) != str(self.port) or str(endport) != str(self.endport):
				continue
			return (tree, ruleset, rule)
		raise exceptions.FirewallError(self, f"Failed to find rule in /etc/vmware/firewall/service.xml")

	def __str__(self):
		return f"<Rule {self.direction} {self.porttype} {self.range} {self.protocol}>"

	def __repr__(self):
		return str(self)


from esxi_utils.file.xml import XmlFile
from esxi_utils.util import log, exceptions
from contextlib import contextmanager
import tempfile
import hashlib
import tarfile
import typing
import shutil
import os
import re

class OvfFile:
	"""
	OVF/OVA file for a pre-configured virtual machine image (virtual appliance), ready to run on a hypervisor. 
	This class provides utility functions for working with an OVF/OVA file, and is agnostic to the actual file format (note: OVA files may have some overhead and require additional disk usage for certain functionality).
	
	:param path: The path to the OVF/OVA file. This may also be a path to a directory containing the file, assuming only one such file exists in that directory (otherwise an exception is raised).
	"""
	def __init__(self, path: str):
		extensions = [".ovf", ".ova"]
		path = os.path.abspath(path)
		if os.path.isdir(path):
			# Is a directory
			# Detect files in this directory
			files = [ os.path.join(path, file) for file in os.listdir(path) if os.path.splitext(file)[1].lower() in extensions ]
			if len(files) == 0:
				raise exceptions.OvfFileNotFoundError(path, f"No OVF/OVA file found")
			if len(files) > 1:
				raise exceptions.OvfAmbiguityError(path, f"Multiple OVF/OVA files exist: {', '.join(files)}")
			path = files[0]
		
		if not os.path.isfile(path):
			raise exceptions.OvfFileNotFoundError(path, f"File does not exist")

		vapp_type = os.path.splitext(path)[1].lower()
		if vapp_type not in extensions:
			raise exceptions.OvfFileNotFoundError(path, f"Path does not point to a valid OVF/OVA file")

		self._path = path
		log.debug(f"{str(self)} Initialized")

	@property
	def path(self) -> str:
		"""
		The path associated with this file.

		:return: The absolute file path.
		"""
		if not os.path.exists(self._path):
			raise exceptions.OvfFileNotFoundError(self.path, f"File does not exist, it may have been moved or deleted")
		return self._path

	@property
	def ext(self) -> str:
		"""
		The extension associated with this file type.

		:return: This file type's extension (either '.ovf' or '.ova').
		"""
		return os.path.splitext(self.path)[1].lower()

	@property
	def is_archive(self) -> bool:
		"""
		Whether or not this is an archive type (i.e. '.ova')

		:return: A boolean
		"""
		return self.ext == ".ova"

	@property
	def descriptor_name(self) -> str:
		"""
		The name of the virtual appliance descriptor file (i.e. the OVF file itself).

		:return: The name of the descriptor file.
		"""
		if not self.is_archive:
			return os.path.basename(self.path)
		ovfs = [ file for file in self._list_files() if file.lower().endswith(".ovf") ]
		if len(ovfs) == 0:
			raise exceptions.OvfFileNotFoundError(self.path, f"Unable to find any OVF files")
		elif len(ovfs) > 1:
			raise exceptions.OvfAmbiguityError(self.path, f"Found several OVF files: {', '.join(ovfs)}")
		return ovfs[0]

	@property
	def manifest_name(self) -> typing.Union[str, None]:
		"""
		The name of the manifest file, if it exists.

		:return: The name of the manifest file, or `None` if it does not exist.
		"""
		descriptor_name = os.path.splitext(os.path.basename(self.descriptor_name))[0]
		for file in self._list_files():
			name, ext = os.path.splitext(file)
			if name == descriptor_name and ext.lower() == ".mf":
				return file
		return None

	@property
	def name(self) -> str:
		"""
		The name of the OVF/OVA.

		:return: A string
		"""
		return os.path.splitext(self.descriptor_name)[0]

	@property
	def vmname(self) -> str:
		"""
		The VM name specified in the OVF/OVA.

		:return: A string
		"""
		return self.descriptor.root["VirtualSystem"][0]["Name"][0].text

	@property
	def ostype(self) -> str:
		"""
		The VM OS type specified in the OVF/OVA.

		:return: A string
		"""
		return self.descriptor.root["VirtualSystem"][0]["OperatingSystemSection"][0].attributes["vmw:osType"]

	@property
	def networks(self) -> typing.List[str]:
		"""
		Get the networks assigned to the OVF/OVA.

		:return: A list of network names.
		"""
		network_sections = self.descriptor.root.find("NetworkSection")
		if len(network_sections) == 0:
			return []
		return [ network.attributes["ovf:name"] for network in network_sections[0]["Network"] ]

	@property
	def files(self) -> typing.List[str]:
		"""
		Get a list of all file names referenced by this virtual appliance (Note: does not check for file existence).

		:return: A list of all file names referenced by this virtual appliance.
		"""
		files = [self.descriptor_name]
		if self.manifest_name:
			files.append(self.manifest_name)
		files.extend([ file.attributes['ovf:href'] for file in self.descriptor.root["References"][0]["File"] ])
		return files

	@property
	def disks(self) -> typing.List[str]:
		"""
		Get a list of all disk names referenced by this virtual appliance (Note: does not check for file existence). This is a subset of `files`.

		:return: A list of all disk names referenced by this virtual appliance.
		"""
		root = self.descriptor.root
		file_id_to_name = { file.attributes['ovf:id']: file.attributes['ovf:href'] for file in root["References"][0]["File"] }
		disks = [ file_id_to_name[disk.attributes['ovf:fileRef']] for disk in root["DiskSection"][0]["Disk"] if disk.attributes['ovf:fileRef'] in file_id_to_name ]
		return disks

	def disk_sizes(self, unit: str = "KB") -> typing.Dict[str, int]:
		"""
		Get the sizes of disk for this virtual machine. 
		This is the maximum capacity of the disk when the OVF/OVA is deployed, not the size of the VMDK as stored on the filesystem.

		:param unit: The unit of measurement to use (one of: B, KB, MB, GB).

		:return: A dictionary mapping disk name to size
		"""
		conversion_factors = { "B": 1, "KB": 1024, "MB": 1024*1024, "GB": 1024*1024*1024 }
		assert unit in conversion_factors
		root = self.descriptor.root
		file_id_to_name = { file.attributes['ovf:id']: file.attributes['ovf:href'] for file in root["References"][0]["File"] }
		sizes = {}
		for disk in root["DiskSection"][0]["Disk"]:
			if disk.attributes['ovf:fileRef'] not in file_id_to_name:
				continue
			name = file_id_to_name[disk.attributes['ovf:fileRef']]
			units = disk.attributes["ovf:capacityAllocationUnits"]
			match = re.match(r"byte \* ([0-9]+)\^([0-9]+)", units)
			if not match:
				raise ValueError(f"Unrecognized capacityAllocationUnits for disk {disk.attributes['ovf:diskId']}: {units}")
			base = int(match.group(1))
			exp =  int(match.group(2))
			capacity = disk.attributes["ovf:capacity"]
			sizes[name] = (int(capacity) * pow(base, exp)) / conversion_factors[unit]
		return sizes

	def required_storage(self, unit: str = "KB") -> int:
		"""
		Get the amount of required storage to contain the VM represented by this virtual appliance. 
		This is the sum of all disk sizes, plus the size of any non-disk files.

		:param unit: The unit of measurement to use (one of: B, KB, MB, GB).

		:return: The storage space required by this OVF/OVA
		"""
		conversion_factors = { "B": 1, "KB": 1024, "MB": 1024*1024, "GB": 1024*1024*1024 }
		assert unit in conversion_factors
		root = self.descriptor.root
		disks = self.disk_sizes(unit="B")
		size = 0
		for file in root["References"][0]["File"]:
			name = file.attributes["ovf:href"]
			if name in disks:
				size += disks[name]
			else:
				size += int(file.attributes["ovf:size"])
		return size / conversion_factors[unit]

	@property
	def descriptor(self) -> 'XmlFile':
		"""
		Open the descriptor file for reading or writing. This returns a class that can be interacted with directly to read or write to the
		underlying descriptor file, and changes will automatically be applied.
		"""
		log.debug(f"{str(self)} Opening descriptor...")

		def update_func(new_xml):
			log.debug(f"{str(self)} Updating descriptor...")
			with self.open(self.descriptor_name, mode="w") as f:
				f.write(new_xml)

		with self.open(self.descriptor_name, mode="r") as f:
			return XmlFile(f, update_func)

	@contextmanager
	def open_descriptor(self) -> typing.Generator['XmlFile', None, None]:
		"""
		Open the descriptor file for reading or writing as a context manager.
		This behaves similarly to the `descriptor` property, but instead applies all changes to the underlying
		OVF at the end of the context for efficiency. This will perform faster when applying multiple changes
		to the descriptor than the `descriptor` property alone. 
		"""
		log.debug(f"{str(self)} Opening descriptor as context manager...")
		with self.open(self.descriptor_name, mode="r+") as f:
			d = XmlFile(f)
			yield d
			f.seek(0)
			f.truncate()
			f.write(d.xml(pretty_print=True, xml_declaration=True))

	@property
	def manifest(self) -> typing.Union[typing.Dict[str, typing.Tuple[str, str]], None]:
		"""
		The contents of the manifest file, as a dict.

		:return: A dict representation of the manifest file, or `None` if a manifest does not exist.
		"""
		if self.manifest_name is None:
			return None

		log.debug(f"{str(self)} Reading manifest...")

		manifest = {}
		with self.open(self.manifest_name, 'r') as f:
			for line in f.readlines():
				line = line.strip()
				if len(line) == 0:
					continue
				match = re.match(r"^(.*?)\((.*?)\)=\s*(.*)$", line)
				manifest[match.group(2)] = (match.group(1), match.group(3))
		return manifest

	@contextmanager
	def open(self, file: str, mode: str = "r") -> typing.Generator[typing.IO, None, None]:
		"""
		Open a file associated with this OVF/OVA.
		Note: For OVAs, this will incur additional overhead as the requested file must be extracted.

		:param file: The name of the file to open.
		:param mode: The mode in which to open the file. 
		"""
		log.info(f"{str(self)} Opening file {file} (mode={mode})")
		if not self.is_archive:
			file = file.strip(os.sep)
			path = os.path.join(os.path.dirname(self.path), file)
			with open(path, mode=mode) as f:
				yield f
			return
		
		# For an archive, we extract the file to a temporary directory and modify that file
		# If this file was modified (detected by hash changes), we'll then re-add that file back to the archive
		file = file.strip(os.sep)
		with tempfile.TemporaryDirectory(prefix="vapp") as tmpdir:
			with tarfile.open(self.path, "r") as tar:
				if file in tar.getnames():
					tar.extract(file, path=tmpdir)

			target = os.path.join(tmpdir, file)

			def get_hash():
				if not os.path.isfile(target):
					return None
				h = hashlib.sha1() # SHA-1 offers fast hashing speed
				with open(target, 'rb') as f:
					while True:
						data = f.read(65536)
						if not data:
							break
						h.update(data)
				return h.hexdigest()

			original_hash = get_hash()
			try:
				with open(target, mode=mode) as f:
					yield f
			finally:
				new_hash = get_hash() 
				log.debug(f"{str(self)} Closing file {file} (mode={mode}). Original hash = {original_hash}, new hash = {new_hash}")
				if new_hash != original_hash:
					log.debug(f"{str(self)} File {file} (mode={mode}) was updated, writing back...")
					# Hash changed, we need to update the original OVA
					# Since we can't update a tar in-place, we'll need to create a new tar archive
					# to replace the original
					with tempfile.TemporaryDirectory(prefix="vapp-tar") as tmptardir:
						new_tar_path = os.path.join(tmptardir, "vapp.tar")
						with tarfile.open(self.path, "r") as original_tar, tarfile.open(new_tar_path, "w", format=tarfile.GNU_FORMAT) as new_tar:
							# Copy files from old archive to the new archive
							for member in original_tar.getmembers():
								if member.name == file:
									# Skip the original file
									continue
								new_tar.addfile(member, fileobj=original_tar.extractfile(member.name))

							# Add the modified file
							new_tar.add(target, arcname=file)

						# Replace the original archive
						path = self.path
						os.remove(path)
						os.rename(new_tar_path, path)
				else:
					log.debug(f"{str(self)} File {file} (mode={mode}) was not updated.")

	def create_manifest(self, hash_type: str = "sha1"):
		"""
		Creates a manifest file, or updates if it already exists.

		:param hash_type: The hash function to use for generating file checksums. Accepts any `hashlib` hash functions.
		"""
		log.info(f"{str(self)} Creating manifest...")

		newContent = ""
		for file in self.files:
			if os.path.splitext(file)[1].lower() in [".mf"]:
				continue

			with self.open(file, mode="rb") as f:
				h = getattr(hashlib, hash_type.lower())()
				while True:
					data = f.read(65536)
					if not data:
						break
					h.update(data)
				digest = h.hexdigest()
				log.debug(f"{str(self)} Hash for {file}: {digest}")
				newContent = newContent + f"{hash_type.upper()}({file})= {digest}\n"
			
		# Write contents
		mf = os.path.splitext(self.descriptor_name)[0] + ".mf"
		with self.open(mf, mode="w") as f:
			f.write(newContent)

	def validate(self):
		"""
		Validates this virtual appliance file by checking that referenced files exist and that the manifest file (if exists) is correct.
		Raises an `OvfManifestError` if the validation fails.
		"""
		log.info(f"{str(self)} Validating...")

		manifest = self.manifest
		def check_manifest_value(f, filename):
			if manifest:
				if filename not in manifest:
					raise exceptions.OvfManifestError(self.path, f"Referenced file \"{filename}\" does not exist in the manifest")
				hashtype = manifest[filename][0]
				expected_digest = manifest[filename][1]
				h = getattr(hashlib, hashtype.lower())()
				while True:
					data = f.read(65536)
					if not data:
						break
					h.update(data)
				digest = h.hexdigest()
				log.debug(f"{str(self)} Expected digest: {expected_digest} (Actual digest: {digest})")
				if digest != expected_digest:
					raise exceptions.OvfManifestError(self.path, f"Manifest does not validate for {filename} (Actual: {digest}; Expected: {expected_digest})")

		with self.open(self.descriptor_name, mode="rb") as f:
			descriptor = XmlFile(f)
			f.seek(0)
			check_manifest_value(f, self.descriptor_name)

		file_ids = []
		for file in descriptor.root["References"][0]["File"]:
			file_ids.append(file.attributes['ovf:id'])
			filename = file.attributes["ovf:href"]

			if filename not in self._list_files():
				raise exceptions.OvfManifestError(self.path, f"Referenced file \"{filename}\" does not exist")

			with self.open(filename, mode="rb") as f:
				size = os.fstat(f.fileno()).st_size
				expected_size = file.attributes['ovf:size']
				log.debug(f"{str(self)} File {filename} expected size: {expected_size} (Actual size: {size})")
				if str(size) != str(expected_size):
					raise exceptions.OvfManifestError(self.path, f"Referenced file \"{filename}\" does not match stated size (Actual {expected_size}; Expected {size})")

				# Check the hash
				check_manifest_value(f, filename)
			
		# Check that disks are referenced properly
		for disk in descriptor.root["DiskSection"][0]["Disk"]:
			file_ref = disk.attributes['ovf:fileRef']
			if file_ref not in file_ids:
				raise exceptions.OvfManifestError(self.path, f"Disk \"{disk.attributes['ovf:fileRef']}\" does not have a reference file.")
				
	def as_ovf(self, path: typing.Optional[str], move: bool = False) -> 'OvfFile':
		"""
		Write this file as an OVF to the provided directory.
		If this is an OVA, the contents will be extracted to the target directory.
		If this is already an OVF, the files will simply be copied to the target directory.

		:param path: The path to the directory in which to place the OVF. If `None`, this will be the same directory containing this file.
		:param move: Whether or not the move the files instead of copying them (delete original files). If `True`, the path for this object will be updated as well.

		:return: An `OvfFile` object for the new OVF.
		"""
		assert isinstance(path, str) or path is None, "path must be a string or None"
		assert isinstance(move, bool), "move must be a boolean"

		log.info(f"{str(self)} Converting to OVF at {path} (move={move})")

		if not self.is_archive and path is None and move == True:
			# Already an OVF, nothing to do
			return self

		if path is None:
			path = os.path.dirname(self.path)

		path = os.path.abspath(path)
		if not os.path.isdir(path):
			raise NotADirectoryError(f"{path} does not refer to an existing directory")

		# Ensure that there are no file conflicts in the target directory
		for file in self.files:
			p = os.path.join(path, file)
			if os.path.exists(p):
				raise FileExistsError(f"{p} already exists")

		# Create the copy
		new_path = os.path.join(path, self.descriptor_name)
		if self.is_archive:
			# Extract the archive
			log.debug(f"{str(self)} Extracting to {new_path}")
			with tarfile.open(self.path, "r") as tar:
				tar.extractall(path)
			if move:
				os.remove(self.path)
		else:
			# Already an OVF, copy the files
			dirname = os.path.dirname(self.path)
			for file in self.files:
				src = os.path.join(dirname, file)
				dst = os.path.join(path, file)
				if os.path.exists(src):
					if move:
						log.debug(f"{str(self)} Moving {src} to {dst}")
						os.rename(src, dst)
					else:
						log.debug(f"{str(self)} Copying {src} to {dst}")
						shutil.copy2(src, dst)

		if move:
			self._path = new_path
		return OvfFile(new_path)

	def as_ova(self, path: typing.Optional[str], move: bool = False) -> 'OvfFile':
		"""
		Write this file as an OVA to the provided directory.
		If this is already an OVA, this will simply be copied to the target directory
		If this is an OVF, the files will be added as a new archive in the target directory.

		:param path: The path to the directory in which to place the OVA. If `None`, this will be the same directory containing this file.
		:param move: Whether or not the move the files instead of copying them (delete original files). If `True`, the path for this object will be updated as well.

		:return: An `OvfFile` object for the new OVA.
		"""
		assert isinstance(path, str) or path is None, "path must be a string"
		assert isinstance(move, bool), "move must be a boolean"

		log.info(f"{str(self)} Converting to OVA at {path} (move={move})")

		if self.is_archive and path is None and move == True:
			# Already an OVA, nothing to do
			return self

		if path is None:
			path = os.path.dirname(self.path)

		path = os.path.abspath(path)
		if not os.path.isdir(path):
			raise NotADirectoryError(f"{path} does not refer to an existing directory")

		# Ensure that there are no file conflicts in the target directory
		new_path = os.path.join(path, f"{self.name}.ova")
		if os.path.exists(new_path):
			raise FileExistsError(f"{new_path} already exists")

		# Create the copy
		if self.is_archive:
			if move:
				log.debug(f"{str(self)} Moving {self.path} to {new_path}")
				os.rename(self.path, new_path)
			else:
				log.debug(f"{str(self)} Copying {self.path} to {new_path}")
				shutil.copy2(self.path, new_path)
		else:
			# This is an OVF, add the files under a new archive
			dirname = os.path.dirname(self.path)
			with tarfile.open(new_path, "w", format=tarfile.GNU_FORMAT) as tar:
				for file in self.files:
					src = os.path.join(dirname, file)
					log.debug(f"{str(self)} Adding {src} to {new_path}")
					tar.add(src, arcname=file)
					if move:
						log.debug(f"{str(self)} Removing {src}")
						os.remove(src)
		
		if move:
			self._path = new_path
		return OvfFile(new_path)

	def remove(self):
		"""
		Delete this file and any associated files.
		"""
		log.info(f"{str(self)} Removing...")

		if self.is_archive:
			log.debug(f"{str(self)} Removing {self.path}")
			os.remove(self.path)
		else:
			dirname = os.path.dirname(self.path)
			for file in self.files:
				path = os.path.join(dirname, file)
				log.debug(f"{str(self)} Removing {path}")
				os.remove(path)

	def rename(self, new_name: str):
		"""
		Rename this virtual appliance and its files. The update will happen in-place.
		Files after the the rename will follow the convention `{new_name}_file{num}` or `{new_name}_disk{num}`
		"""
		assert isinstance(new_name, str), "new_name must be a string"

		log.info(f"{str(self)} Renaming to {new_name}...")

		with tempfile.TemporaryDirectory(prefix="vapp-rename") as tmpdir:
			# In order to avoid having two separate functions for renaming OVF and OVAs, we'll create a common API
			# by working with the raw files in a temporary directory. If this is an OVF, we simply copy the files to
			# the temporary directory, and if this is an OVA we extract the files instead.
			# Before exiting this function, we'll replace the original file(s)
			# This has the added benefit of ensuring that the original files are not touched until the very end
			# i.e. the original files won't be modified if this function fails
			tmpovf = self.as_ovf(tmpdir, move=False)
			create_manifest = False
			if tmpovf.manifest_name:
				# Remove the manifest as it will become invalid
				# We'll re-create it later
				create_manifest = True
				temp_manifest_path = os.path.join(tmpdir, tmpovf.manifest_name)
				log.debug(f"{str(self)} Removing {temp_manifest_path}...")
				os.remove(temp_manifest_path)

			with open(os.path.join(tmpdir, tmpovf.descriptor_name)) as f:
				descriptor = XmlFile(f)

			old_value = descriptor.root["VirtualSystem"][0].attributes['ovf:id']
			log.debug(f"{str(self)} Setting VirtualSystem ID ({old_value}) to {new_name}")
			descriptor.root["VirtualSystem"][0].attributes['ovf:id'] = new_name

			old_value = descriptor.root["VirtualSystem"][0]["Name"][0].text
			log.debug(f"{str(self)} Setting VirtualSystem Name ({old_value}) to {new_name}")
			descriptor.root["VirtualSystem"][0]["Name"][0].text = new_name

			old_value = descriptor.root["VirtualSystem"][0]["VirtualHardwareSection"][0]["System"][0]["vssd:VirtualSystemIdentifier"][0].text
			log.debug(f"{str(self)} Setting VirtualSystemIdentifier ({old_value}) to {new_name}")
			descriptor.root["VirtualSystem"][0]["VirtualHardwareSection"][0]["System"][0]["vssd:VirtualSystemIdentifier"][0].text = new_name
			
			# Update files
			# The standard naming convention is {Name}_file{num} or {Name}_disk{num}
			# The rename will enforce this naming convention, even if the original files did not follow it
			num_files = 0
			num_disks = 0
			disk_file_ids = [ disk["fileRef"] for disk in descriptor.root["DiskSection"][0]["Disk"] ]
			for file in descriptor.root["References"][0]["File"]:
				# Update href
				old_href = file.attributes["ovf:href"]
				extension = os.path.splitext(old_href)[1]
				new_href = ""
				if file.attributes["ovf:id"] in disk_file_ids:
					# This is a disk
					new_href = f"{new_name}_disk{num_disks+1}{extension}"
					num_disks += 1
				else:
					new_href = f"{new_name}_file{num_files+1}{extension}"
					num_files += 1

				log.debug(f"{str(self)} Setting file {file.attributes['ovf:id']} href {file.attributes['ovf:href']} to {new_href}")
				file.attributes["ovf:href"] = new_href

				src = os.path.join(tmpdir, old_href)
				dst = os.path.join(tmpdir, new_href)
				log.debug(f"{str(self)} Renaming {src} to {dst}")
				os.rename(src, dst)
			
			# Write the descriptor
			temp_descriptor_path = os.path.join(tmpdir, tmpovf.descriptor_name)
			log.debug(f"{str(self)} Writing new descriptor to {temp_descriptor_path}")
			with open(temp_descriptor_path, "w") as f:
				f.write(descriptor.xml(pretty_print=True, xml_declaration=True))

			# Rename the OVF file itself
			src = os.path.join(tmpdir, tmpovf.descriptor_name)
			dst = os.path.join(tmpdir, f"{new_name}.ovf")
			log.debug(f"{str(self)} Renaming descriptor {src} to {dst}")
			os.rename(src, dst)
			tmpovf = OvfFile(dst)

			# If the manifest existed re-create it under the correct name
			# Note: this has to be done last to ensure the hashes are correct
			if create_manifest:
				tmpovf.create_manifest()

			# Move the modified files to the correct location and delete the originals
			dirname = os.path.dirname(self.path)
			if self.is_archive:
				tmpovf.as_ova(dirname, move=True)
			else:
				tmpovf.as_ovf(dirname, move=True)
			self.remove()
			self._path = tmpovf.path

	def rename_network(self, old_network_name: str, new_network_name: str):
		"""
		Rename a network in this OVF or OVA. The update will happen in-place.
		If the network is not found in the OVF, an exception will be raised.

		:param old_network_name: The name of the old network to update.
		:param new_network_name: The new name of the network.
		"""
		assert isinstance(old_network_name, str), "old_network_name must be a string"
		assert isinstance(new_network_name, str), "new_network_name must be a string"

		log.info(f"{str(self)} Renaming network {old_network_name} to {new_network_name}...")
		with self.open_descriptor() as descriptor:
			if len(descriptor.root["NetworkSection"]) == 0:
				raise exceptions.OvfFileError(self.path, f"OVF file does not contain a network section")

			for ovf_network in descriptor.root["NetworkSection"][0]["Network"]:
				old_value =  ovf_network.attributes["ovf:name"]
				if old_value != old_network_name:
					continue

				log.debug(f"{str(self)} Setting Network Name ({old_value}) to {new_network_name}")
				ovf_network.attributes["ovf:name"] = new_network_name

				old_value = ovf_network["Description"][0].text
				new_value = ovf_network["Description"][0].text.replace(old_network_name, new_network_name)
				log.debug(f"{str(self)} Setting Network Description ({old_value}) to {new_value}")
				ovf_network["Description"][0].text = new_value
				break
			else:
				raise exceptions.OvfFileError(self.path, f"Could not find network \"{old_network_name}\"")

			# Update network hardware
			for hardware_item in descriptor.root["VirtualSystem"][0]["VirtualHardwareSection"][0]["Item"]:
				if hardware_item["rasd:ResourceType"][0].text.strip() not in ["10", "11"]: # 10 = Ethernet Adapter; 11 = Other Network Adapter
					continue
				old_value = hardware_item["rasd:Connection"][0].text
				if old_value != old_network_name:
					continue

				log.debug(f"{str(self)} Setting Connection ({old_value}) to {new_network_name}")
				hardware_item["rasd:Connection"][0].text = new_network_name

				old_value = hardware_item["rasd:Description"][0].text
				new_value = hardware_item["rasd:Description"][0].text.replace(old_network_name, new_network_name)
				log.debug(f"{str(self)} Setting Connection Description ({old_value}) to {new_network_name}")
				hardware_item["rasd:Description"][0].text = new_value
				break
			else:
				raise exceptions.OvfFileError(self.path, f"Could not find network \"{old_network_name}\"")

		if self.manifest_name:
			# Fix the manifest
			self.create_manifest()

	def set_config(self, key: str, value: typing.Union[str, bool], required: bool = False, extraconfig=False):
		"""
		Set a config entry (a <Config> tag) in the OVF. If the entry does not exist, it will be created.

		:param key: The entry key.
		:param value: The entry value.
		:param required: Whether or not the entry is set to be required.
		:param extraconfig: Whether or not this is an extra-config entry (an <ExtraConfig> tag) rather than a standard config entry.
		"""
		if isinstance(value, bool):
			value = "true" if value else "false"
		tag = "vmw:ExtraConfig" if extraconfig else "vmw:Config"

		root = self.descriptor.root["VirtualSystem"][0]["VirtualHardwareSection"][0]
		matching = [config for config in root[tag] if config.attributes["vmw:key"] == key ]

		config = None
		if len(matching) != 0:
			config = matching[0]
		else:
			config = root.append(tag)
		config.attributes["ovf:required"] = "true" if required else "false"
		config.attributes["vmw:key"] = key
		config.attributes["vmw:value"] = value

	def remove_config(self, key: str, extraconfig=False):
		"""
		Remove a config entry (a <Config> tag) in the OVF. If the entry does not exist, this does nothing.

		:param key: The entry key.
		:param extraconfig: Whether or not this is an extra-config entry (an <ExtraConfig> tag) rather than a standard config entry.
		"""
		tag = "vmw:ExtraConfig" if extraconfig else "vmw:Config"
		entries = self.descriptor.root["VirtualSystem"][0]["VirtualHardwareSection"][0][tag]
		for config in entries:
			if config.attributes["vmw:key"] == key:
				config.remove()

	def _list_files(self):
		"""
		List files in this OVF's parent directory, or a list of files contained in the OVA archive.

		:return: A list of file names.
		"""
		if not self.is_archive:
			return os.listdir(os.path.dirname(self.path))
		with tarfile.open(self.path, "r") as tar:
			return tar.getnames()

	def __str__(self):
		return f"<{type(self).__name__} at {self._path}>"

	def __repr__(self):
		return str(self)
		


from esxi_utils.firewall.ruleset import RulesetList
from esxi_utils.util import log
import concurrent.futures
from lxml import etree
import typing
import socket
import re

if typing.TYPE_CHECKING:
	from esxi_utils.client import ESXiClient

class Firewall:
	"""
	The host firewall.
	"""
	def __init__(self, client: 'ESXiClient'):
		self._client = client

	@property
	def default_policy(self) -> typing.Dict[str, bool]:
		"""
		The default firewall policy. A dict containing the keys `IncomingBlocked` and `OutgoingBlocked`
		"""
		return {"IncomingBlocked": self._firewall_info.defaultPolicy.incomingBlocked, "OutgoingBlocked": self._firewall_info.defaultPolicy.outgoingBlocked}

	@property 
	def rulesets(self) -> 'RulesetList':
		"""
		Rulesets for the firewall.
		"""
		return RulesetList(self)

	def find_vnc_port(self) -> typing.List[int]:
		"""
		Detect an open VNC port on the ESXi host. VNC ports must be used by one VM only, and must be exposed by firewall rules (ports 5900-6000).
		This function will attempt to detect available ports by sequentially connecting to each within the VNC port range. 
		If a port is found, it is advised to assign the port to a VM immediately to avoid another VM from acquiring the same port.

		:return: A list of ints (valid VNC ports)
		"""
		log.info(f"Searching for VNC port...")

		# This process is derived from the process used by the Packer VMware plugin
		# Process ports ESXi is listening on to determine which are available
		# This process does best effort to detect ports that are unavailable,
		# it will ignore any ports listened to by only localhost
		with self._client.ssh() as conn:
			ip_data_list = conn.esxcli("network ip connection list")
		listen_ports = dict()
		for record in ip_data_list:
			if record["State"] == "LISTEN":
				split_address = record["LocalAddress"].split(":")
				if split_address[0] != "127.0.0.1":
					listen_ports[int(split_address[-1])] = True
		
		def thread_func(test_port):
			nonlocal listen_ports
			if test_port in listen_ports:
				log.debug(f"VNC Port {test_port} in use.")
				return None # In use
			s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			s.settimeout(10)
			try:
				s.connect((self._client.hostname, test_port))
			except Exception as e:
				if "timed out" not in str(e).lower():
					log.debug(f"VNC Port {test_port} valid.")
					return test_port
				else:
					log.debug(f"VNC Port {test_port} unavailable ({str(e)}).")
			finally:
				s.close()
			return None
		
		valid_ports = []
		with concurrent.futures.ThreadPoolExecutor(max_workers=100) as executor:
			futures = []
			for port in range(5900, 6001):
				futures.append(executor.submit(thread_func, port))
			for future in concurrent.futures.as_completed(futures):
				valid = future.result()
				if valid:
					valid_ports.append(valid)
			
		return valid_ports
		
	def _get_service_xml(self):
		"""
		Get the contents of /etc/vmware/firewall/service.xml as an `etree` object.
		"""
		with self._client.ssh() as conn:
			service_xml = conn.read("/etc/vmware/firewall/service.xml")
		return etree.fromstring(service_xml, parser=etree.XMLParser(remove_blank_text=True))

	def _update_service_xml(self, new_tree):
		"""
		Update /etc/vmware/firewall/service.xml with new content

		:param new_tree: The `etree` XML object to write to the service file.
		"""
		new_content = etree.tostring(new_tree, pretty_print=True, encoding='unicode')
		new_content = re.sub(r"</service>", "\g<0>\n", new_content, flags=re.MULTILINE)
		with self._client.ssh() as conn:
			err = None
			try:
				# See: https://kb.vmware.com/s/article/2008226
				conn.exec("chmod 644 /etc/vmware/firewall/service.xml")
				conn.exec("chmod +t /etc/vmware/firewall/service.xml")
				conn.write("/etc/vmware/firewall/service.xml", new_content, overwrite=True)
				conn.exec("esxcli network firewall refresh", assert_status=0)
			except Exception as e:
				err = e
			finally:
				conn.exec("chmod 444 /etc/vmware/firewall/service.xml")
				conn.exec("chmod +t /etc/vmware/firewall/service.xml")
			if err:
				raise err

	@property
	def _firewall_system(self):
		return self._client._host_system.configManager.firewallSystem

	@property
	def _firewall_info(self):
		return self._firewall_system.firewallInfo

	def __str__(self):
		return f"<{type(self).__name__} for {self._client.hostname}>"

	def __repr__(self):
		return str(self)


from esxi_utils.vm.virtualmachine import VirtualMachine
from esxi_utils.vm.types.ostype import OSType
from esxi_utils.vm.types.cisco import CiscoVirtualMachine
from esxi_utils.vm.types.linux import LinuxVirtualMachine
from esxi_utils.vm.types.panos import PaloAltoFirewallVirtualMachine
from esxi_utils.vm.types.windows import WindowsVirtualMachine

from esxi_utils.vm import hardware

__all__ = [
	"VirtualMachine",
	"OSType",
	"CiscoVirtualMachine",
	"LinuxVirtualMachine",
	"PaloAltoFirewallVirtualMachine",
	"WindowsVirtualMachine"
]


from esxi_utils.vm.virtualmachine import VirtualMachine
from esxi_utils.util.connect.cisco import CiscoSSHConnection
import typing

if typing.TYPE_CHECKING:
	from esxi_utils.vm.types.ostype import OSType

class CiscoVirtualMachine(VirtualMachine):
	"""
	A Cisco virtual machine on an ESXi host. 
	"""
	@property
	def ssh(self) -> typing.Type['CiscoSSHConnection']:
		"""
		A SSH connection object for establishing a remote connection to this Cisco VM.
		"""
		return CiscoSSHConnection

	@property
	def ostype(self) -> 'OSType':
		"""
		The `OSType` for this Virtual Machine.
		"""
		from esxi_utils.vm.types.ostype import OSType
		return OSType.Cisco
		


from esxi_utils.vm.hardware.device import VirtualDevice, VirtualDeviceList
from esxi_utils.util import log
import pyVmomi
import typing


class VirtualVideoCardList(VirtualDeviceList):
	"""
	The list of all virtual video cards on a Virtual Machine.
	"""
	def __iter__(self) -> typing.Iterator['VirtualVideoCard']:
		cards = [ dev for dev in super().__iter__() if isinstance(dev, VirtualVideoCard) ]
		# cards.sort(key=lambda x: int(re.search(r"\d+", x.label).group(0)))
		return iter(cards)

	def __getitem__(self, index) -> 'VirtualVideoCard':
		return self.items[index]

	@property
	def items(self) -> 'VirtualVideoCard':
		"""
		A list of all items
		"""
		return list(self)


class VirtualVideoCard(VirtualDevice):
    @property
    def videoRamSizeKB(self) -> int:
        """
        The current size of the video card VRAM
        """
        return self._obj.videoRamSizeInKB
	
    @property
    def graphicsMemorySizeKB(self) -> int:
        """
        The current size of the 3D graphics memory
        """
        return self._obj.graphicsMemorySizeInKB
    
    @property
    def enable3D(self) -> bool:
        """
        Whether or not (3D) graphics are currently enabled on this video card
        """
        return self._obj.enable3DSupport
    
    @property
    def use3Drenderer(self) -> str:
        """
        Returns a string such as 'automatic'
        """
        return self._obj.use3dRenderer
    
    @property
    def useAutoDetect(self) -> bool:
        """
        When true: VM will auto-detect graphic settings from the VM. When false: depends on custom settings as set by the user in the GUI
        """
        return self._obj.useAutoDetect

    @property
    def numDisplays(self) -> int:
        """
        The number of displays connected to this video card
        """
        return self._obj.numDisplays
    
    def createVideoCardEditSpec(self) -> pyVmomi.vim.vm.device.VirtualDeviceSpec:
        """
        The boilerplate template object that will be passed to 'edit'/reconfigure the video card.
        All video card setter tasks will need this base object.

        :returns: the pyVmomi object needed to reconfigure the video card
        """
        video_card_spec = pyVmomi.vim.vm.device.VirtualDeviceSpec()
        video_card_spec.operation = pyVmomi.vim.vm.device.VirtualDeviceSpec.Operation.edit
        video_card_spec.device = self._obj
        return video_card_spec

    def videoCardReconfigureTask(self, video_card_spec: pyVmomi.vim.vm.device.VirtualDeviceSpec):
        """
        Nests the video_card_spec inside the generate configuration spec needed by pyVmomi. Executes the video card reconfiguration/'edit' task. Waits for the task to finish.
        """
        spec = pyVmomi.vim.vm.ConfigSpec()
        spec.deviceChange = [video_card_spec]
        self._vm._client._wait_for_task(
            self._vm._vim_vm.ReconfigVM_Task(spec=spec)
        )
    
    @useAutoDetect.setter
    def useAutoDetect(self, use: bool):
        """
        Set the auto detect flag for this video card.

        :param use: When true: VM will auto-detect graphic settings from the VM. When false: depends on custom settings as set by the user in the GUI
        """
        log.info(f"{str(self)} Updating auto-detect to: {use}")
        
        video_card_spec = self.createVideoCardEditSpec()
        video_card_spec.device.useAutoDetect = use
        self.videoCardReconfigureTask(video_card_spec)

    @graphicsMemorySizeKB.setter
    def graphicsMemorySizeKB(self, size: int):
        """
        Set the (3D) graphics memory size for this video card. An error may occur if the size isn't supported by the VM or ESXi itself.
        The VM must be turned off to change this setting.
        THIS VALUE IS NOT THE ONE SHOWN IN THE ESXI UI (see 'videoRamSizeKB' for the VRAM value shown as 'Total Video Memory' in the UI.)
        You should only try to set the size of the video card if 'useAutoDetect' is false.

        :param size: The size in KB (KiloBytes) to set the video card 'graphics' memory to (e.g. 1024 KB is 1 MB)
        """
        log.info(f"{str(self)} Updating graphics memory size to: {size}")
        
        video_card_spec = self.createVideoCardEditSpec()
        video_card_spec.device.graphicsMemorySizeInKB = size
        self.videoCardReconfigureTask(video_card_spec)

    @videoRamSizeKB.setter
    def videoRamSizeKB(self, size: int):
        """
        Set the VRAM size for this video card. An error may occur if the size isn't supported by the VM or ESXi itself. This is the 'Total Video Memory' field as shown in the ESXi UI.
        The VM must be turned off to change this setting.
        You should only try to set the size of the video card if 'useAutoDetect' is false.

        :param size: The size in KB (KiloBytes) to set the video card VRAM to (e.g. 1024 KB is 1 MB)
        """
        log.info(f"{str(self)} Updating VRAM size to: {size}")
        
        video_card_spec = self.createVideoCardEditSpec()
        video_card_spec.device.videoRamSizeInKB = size
        self.videoCardReconfigureTask(video_card_spec)

    @enable3D.setter
    def enable3D(self, enabled: bool):
        """
        Set 3D support on the video card to be either enabled or disabled.
        The VM must be turned off to change this setting.
        You should only try to change this value if the video card 'useAutoDetect' field is false.

        :param enabled: Whether to set 3D to enabled on this video card or not
        """
        log.info(f"{str(self)} Updating 3D status to: {enabled}")
        
        video_card_spec = self.createVideoCardEditSpec()
        video_card_spec.device.enable3DSupport = enabled
        self.videoCardReconfigureTask(video_card_spec)

    @use3Drenderer.setter
    def use3Drenderer(self, renderer_name: str):
        """
        Set the name of the 3D renderer to use.
        You should only try to change this value if the video card 'useAutoDetect' field is false.

        :param renderer_name: The name of the renderer to use (options appear to be: Automatic, Hardware, Software)
        """
        log.info(f"{str(self)} Updating 3D renderer to: {renderer_name}")
        
        video_card_spec = self.createVideoCardEditSpec()
        video_card_spec.device.use3Drenderer = renderer_name
        self.videoCardReconfigureTask(video_card_spec)

    @numDisplays.setter
    def numDisplays(self, amount: int):
        """
        Set the number of displays connected to this video card.
        You should only try to set the size of the video card if 'useAutoDetect' is false.

        :param amount: the number of displays to connect to this video card (options in the UI are 1 through 10)
        """
        log.info(f"{str(self)} Updating number of displays to: {amount}")
        
        video_card_spec = self.createVideoCardEditSpec()
        video_card_spec.device.numDisplays = amount
        self.videoCardReconfigureTask(video_card_spec)

    def __str__(self):
        if self.useAutoDetect:
            return f"<{type(self).__name__} with auto-detect video settings set to 'True' for VM='{self._vm.name}'>"
        return f"<{type(self).__name__} with custom VRAM setting: '{self.videoRamSizeKB} KB' for VM='{self._vm.name}'>" 



from esxi_utils.util.connect.ssh import SSHConnection, SSHResponsePromise
from dateutil.parser import parse as parsetime
from esxi_utils.util import log, exceptions
import datetime
import typing
import uuid
import stat
import os
import re

if typing.TYPE_CHECKING:
	from esxi_utils.util.response import Response

class UnixSSHConnection(SSHConnection):
	"""
	A SSH connection to a Unix or Unix-like remote host.

	:param ip: The IP of the target.
	:param username: The username to use when logging into the remote system.
	:param password: The password to use when logging into the remote system.
	:param sudo_password: The password to use when running as sudo. Only necessary if the sudo password differs from `password`, as `password` will be used as the default.
	:param port: The port to use when connecting over SSH.
	:param min_pty_width: The minimum width of pseudoterminals (in columns) when ``pty=True`` is set on a command.
	:param min_pty_height: The minimum height of pseudoterminals (in rows) when ``pty=True`` is set on a command.
	"""
	def __init__(self, ip: str, username: str, password: str, sudo_password: typing.Optional[str] = None, port: int = 22, min_pty_width: int = 160, min_pty_height: int = 48):
		super().__init__(
			ip=ip, 
			username=username,
			password=password,
			port=port,
			min_pty_width=min_pty_width,
			min_pty_height=min_pty_height,
		)
		self._sudo_password = sudo_password

	def exec(
		self, 
		cmd: str, 
		stdin: typing.Optional[typing.Dict[str, str]] = None, 
		timeout: int = 120, 
		pty: bool = False, 
		cwd: typing.Optional[str] = None, 
		env: typing.Optional[typing.Dict[str, typing.Any]] = None, 
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None,
		assert_status: typing.Optional[int] = None,
		network_retries: int = 0,
		sudo_password: typing.Optional[str] = None,
	) -> 'Response':
		"""
		Execute a shell command on the remote end of this connection. Includes some Unix-specific functionality, such as handling sudo prompts (hint: use `sudo {cmd}` to run a command as root).

		:param cmd: The command to run.
		:param stdin: A dict mapping a 'pattern' (regex string) to a 'response' (string), to input 'response' to stdin when stdout matches 'pattern'
		:param timeout: The command timeout in seconds. Set to 0 to disable timeout.
		:param pty: Run the command in a pseudoterminal. If `True`, the returned `Response` object will have a merged stdout and stderr, and stderr will be empty.
		:param cwd: The directory that the command should run in. If `None`, runs in the default directory.
		:param env: A dict of shell environment variables to be merged into the default environment that the remote command executes within.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.
		:param network_retries: Number of times to retry the command when a recoverable network error is encountered. The command should be idempotent when this is non-zero as it may run several times.
		:param sudo_password: The password to use for sudo prompts. If not provided, this will default first to the `sudo_password` on the SSH connection object, and then to the `password` on the SSH connection object.
		
		:return: A `Response` object.
		"""
		if not isinstance(stdin, dict):
			stdin = dict()
		stdin = { key: value for key, value in stdin.items() } # Create a copy
		sudo_pass = self._password
		if sudo_password:
			sudo_pass = sudo_password
		elif self._sudo_password:
			sudo_pass = self._sudo_password
		stdin[r"\[sudo\] password"] = sudo_pass + "\n"
		return SSHConnection.exec(
			self,
			cmd=cmd,
			stdin=stdin,
			timeout=timeout,
			pty=pty,
			cwd=cwd,
			env=env,
			out_stream_callback=out_stream_callback,
			assert_status=assert_status,
			network_retries=network_retries
		)
	
	def exec_script(
		self, 
		script: str, 
		stdin: typing.Optional[typing.Dict[str, str]] = None, 
		timeout: int = 120, 
		pty: bool = False, 
		cwd: typing.Optional[str] = None, 
		env: typing.Optional[typing.Dict[str, typing.Any]] = None, 
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None,
		assert_status: typing.Optional[int] = None,
		network_retries: int = 0,
		use_sudo: bool = False,
		sudo_password: typing.Optional[str] = None,
	) -> 'Response':
		"""
		Execute a script on the remote end of this connection. This will write a temporary script file and execute it on the remote end of this connection.

		:param script: The script to run.
		:param stdin: A dict mapping a 'pattern' (regex string) to a 'response' (string), to input 'response' to stdin when stdout matches 'pattern'
		:param timeout: The command timeout in seconds. Set to 0 to disable timeout.
		:param pty: Run the command in a pseudoterminal. If `True`, the returned `Response` object will have a merged stdout and stderr, and stderr will be empty.
		:param cwd: The directory that the command should run in. If `None`, runs in the default directory.
		:param env: A dict of shell environment variables to be merged into the default environment that the remote command executes within.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.
		:param network_retries: Number of times to retry the command when a recoverable network error is encountered. The script should be idempotent when this is non-zero as it may run several times.
		:param use_sudo: Whether to run the entire script using sudo. Setting this to `False` does not forgo the possibility of using sudo commands within the script itself.
		:param sudo_password: The password to use for sudo prompts. If not provided, this will default first to the `sudo_password` on the SSH connection object, and then to the `password` on the SSH connection object.
		
		:return: A `Response` object.
		"""
		script = re.sub(r"^\s*", "", script, flags=re.MULTILINE).strip()
		script_path = f"/tmp/{str(uuid.uuid4()).replace('-', '')}.sh"
		self.write(script_path, script)
		
		script_cmd = f"bash -eu {script_path}"
		if use_sudo:
			script_cmd = "sudo " + script_cmd
		try:
			resp = self.exec(
				cmd=script_cmd,
				stdin=stdin,
				timeout=timeout,
				pty=pty,
				cwd=cwd,
				env=env,
				out_stream_callback=out_stream_callback,
				assert_status=assert_status,
				network_retries=network_retries,
				sudo_password=sudo_password
			)
			resp.cmd = script
			return resp
		finally:
			self.rm(script_path)

	def exec_async(
		self, 
		cmd: str, 
		stdin: typing.Optional[typing.Dict[str, str]] = None, 
		timeout: int = 120, 
		pty: bool = False,
		cwd: typing.Optional[str] = None, 
		env: typing.Optional[typing.Dict[str, typing.Any]] = None, 
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None
	) -> 'SSHResponsePromise':
		"""
		Execute a shell command asynchronously on the remote end of this connection. Includes some Unix-specific functionality, such as handling sudo prompts (hint: use `sudo {cmd}` to run a command as root).

		:param cmd: The command to run.
		:param stdin: A dict mapping a 'pattern' (regex string) to a 'response' (string), to input 'response' to stdin when stdout matches 'pattern'
		:param timeout: The command timeout in seconds. Set to 0 to disable timeout.
		:param pty: Run the command in a pseudoterminal. If `True`, the returned `Response` object will have a merged stdout and stderr, and stderr will be empty.
		:param cwd: The directory that the command should run in. If `None`, runs in the default directory.
		:param env: A dict of shell environment variables to be merged into the default environment that the remote command executes within.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 

		:return: A `SSHResponsePromise` object.
		"""
		if not isinstance(stdin, dict):
			stdin = dict()
		stdin = { key: value for key, value in stdin.items() } # Create a copy
		sudo_pass = self._password
		if self._sudo_password:
			sudo_pass = self._sudo_password
		stdin[r"\[sudo\] password"] = sudo_pass + "\n"
		return SSHConnection.exec_async(
			self,
			cmd=cmd,
			stdin=stdin,
			timeout=timeout,
			pty=pty,
			cwd=cwd,
			env=env,
			out_stream_callback=out_stream_callback
		)

	def stat(self, path: str, cwd: typing.Optional[str] = None) -> typing.Union[typing.Dict[str, typing.Any], None]:
		"""
		Get information about a file.

		:param path: A path to a file or directory on the remote system.
		:param cwd: The current working directory. If `None`, no current working directory is used.
		
		:return: A dict containing information about the file, or `None` if the file does not exist.
		"""
		log.debug(f"{str(self)} Stat {path}")
		assert self._connection and self._connection.client, "Connection is not open"
		with self._connection.client.open_sftp() as ftp:
			ftp.chdir(cwd)
			try:
				s = ftp.stat(path)
			except FileNotFoundError:
				return None
			return {
				"name": os.path.basename(path),
				"size": s.st_size,
				"mode": s.st_mode,
				"isfile": stat.S_ISREG(s.st_mode), # type: ignore
				"isdir": stat.S_ISDIR(s.st_mode), # type: ignore
				"mtime": s.st_mtime,
				"atime": s.st_atime,
				"gid": s.st_gid,
				"uid": s.st_uid
			}

	def isfile(self, path: str, cwd: typing.Optional[str] = None) -> bool:
		"""
		Checks if the file exists and is a regular file.

		:param path: A path to a file on the remote system.
		:param cwd: The current working directory. If `None`, no current working directory is used.
		
		:return: A boolean whether or not the file exists.
		"""
		stat = self.stat(path, cwd=cwd)
		if stat is None:
			return False
		return stat["isfile"]

	def isdir(self, path: str, cwd: typing.Optional[str] = None) -> bool:
		"""
		Checks if the directory exists.

		:param path: A path to a directory on the remote system.
		:param cwd: The current working directory. If `None`, no current working directory is used.
		
		:return: A boolean whether or not the directory exists.
		"""
		stat = self.stat(path, cwd=cwd)
		if stat is None:
			return False
		return stat["isdir"]

	def touch(self, path: str, cwd: typing.Optional[str] = None):
		"""
		Create an empty file.

		:param path: Path to the new file on the remote system.
		:param cwd: The current working directory. If `None`, no current working directory is used.
		"""
		self.exec(f"touch {path}", assert_status=0, cwd=cwd)

	def mkdir(self, path: str, parents: bool = False, cwd: typing.Optional[str] = None):
		"""
		Create a directory.

		:param path: Path to the new directory on the remote system.
		:param parents: Whether or not the create parent directories as needed.
		:param cwd: The current working directory. If `None`, no current working directory is used.
		"""
		self.exec(f"mkdir -p {path}" if parents else f"mkdir {path}", assert_status=0, cwd=cwd)
	
	def rm(self, path: str, recursive: bool = False, cwd: typing.Optional[str] = None):
		"""
		Remove a file or directory.

		:param path: Path to the directory to remove on the remote system.
		:param recursive: Whether or not the delete directories recursively (required to be True for removing directories).
		:param cwd: The current working directory. If `None`, no current working directory is used.
		"""
		self.exec(f"rm -rf {path}" if recursive else f"rm -f {path}", assert_status=0, cwd=cwd)

	def cp(self, src: str, dst: str, recursive: bool = False, cwd: typing.Optional[str] = None):
		"""
		Copy a file or directory.

		:param src: The source for the `cp` command. File globs may be used.
		:param dst: The destination for the `cp` command. If the destination exists, it will be overwritten.
		:param recursive: Copy directories and their contents recursively (required to be True for copying directories).
		:param cwd: The current working directory. If `None`, no current working directory is used.
		"""
		self.exec(f"cp -r {src} {dst}" if recursive else f"cp {src} {dst}", assert_status=0, cwd=cwd)

	def mv(self, src: str, dst: str, cwd: typing.Optional[str] = None):
		"""
		Move a file or directory.

		:param src: The source for the `mv` command. File globs may be used.
		:param dst: The destination for the `mv` command. If the destination exists, it will be overwritten.
		:param cwd: The current working directory. If `None`, no current working directory is used.
		"""
		self.exec(f"mv -f {src} {dst}", assert_status=0, cwd=cwd)

	def ls(self, path: str, cwd: typing.Optional[str] = None, sudo_dash_S: bool=False, sudo_password: typing.Optional[str] = None):
		"""
		List the files within a directory using SFTP.

		:param path: Path to the directory on the remote system.
		:param cwd: The current working directory. If `None`, no current working directory is used.
		:param sudo_dash_S: when 'True' use exec the command: 'sudo -S ls -a' instead of using ftp.listdir() (cwd will be ignored)
		:param sudo_password: The password to use for sudo prompts. If not provided, this will default first to the `sudo_password` on the SSH connection object, and then to the `password` on the SSH connection object. This parameter is only used when sudo_dash_S is provided.
		"""
		assert self._connection and self._connection.client, "Connection is not open"
		if sudo_dash_S:
			return self.exec(f'sudo -S ls -a {path}', sudo_password=sudo_password).stdout.split('\n')
		with self._connection.client.open_sftp() as ftp:
			ftp.chdir(cwd)
			return ftp.listdir(path)

	def read(self, path: str, encoding: typing.Optional[str] = 'utf-8', cwd: typing.Optional[str] = None, sudo_dash_S: bool=False, sudo_password: typing.Optional[str] = None) -> typing.Union[str, bytes]:
		"""
		Read the contents of a file using SFTP.

		:param path: Path to the file on the remote system.
		:param encoding: The encoding to use when decoding the file to string. If `None`, no decoding is performed and bytes will be returned.
		:param cwd: The current working directory. If `None`, no current working directory is used.
		:param sudo_dash_S: when 'True' use exec the command: 'sudo -S cat' instead of using ftp.file.read() (cwd will be ignored)
		:param sudo_password: The password to use for sudo prompts. If not provided, this will default first to the `sudo_password` on the SSH connection object, and then to the `password` on the SSH connection object. This parameter is only used when sudo_dash_S is provided.
		
		:return: The contents of the remote file.
		"""
		assert self._connection and self._connection.client, "Connection is not open"
		if sudo_dash_S:
			return self.exec(f'sudo -S cat {path}', sudo_password=sudo_password).stdout
		with self._connection.client.open_sftp() as ftp:
			ftp.chdir(cwd)
			file = ftp.file(path, "r")
			b = file.read()
			file.close()
			if encoding is not None:
				return b.decode(encoding)
			return b
	
	def write(self, path: str, contents: typing.Union[str, bytes], overwrite: bool = False, cwd: typing.Optional[str] = None, sudo_dash_S: bool=False, sudo_password: typing.Optional[str] = None):
		"""
		Write to a file using SFTP. The file will be created if it does not exist.

		:param path: Path to the file on the remote system.
		:param contents: The contents to write to the file.
		:param overwrite: Whether or not to overwrite the file's existing contents. If false, the contents will be appended instead.
		:param cwd: The current working directory. If `None`, no current working directory is used.
		:param sudo_dash_S: when 'True' use exec the command: 'sudo -S echo _contents_ > _path_ ' instead of using ftp.file.write() (cwd will be ignored)
		:param sudo_password: The password to use for sudo prompts. If not provided, this will default first to the `sudo_password` on the SSH connection object, and then to the `password` on the SSH connection object. This parameter is only used when sudo_dash_S is provided.
		"""
		assert self._connection and self._connection.client, "Connection is not open"
		# exec
		if sudo_dash_S:
			redirect_char = '>' if overwrite else '>>'
			return self.exec(f'sudo -S sh -c \'echo "{contents}" {redirect_char} {path}\'', sudo_password=sudo_password)

		# sftp
		mode = "a"
		if overwrite:
			mode = "w"
		with self._connection.client.open_sftp() as ftp:
			ftp.chdir(cwd)
			file = ftp.file(path, mode)
			file.write(contents)
			file.close()

	def download(self, path: str, dst: str, directory_contents_only: bool = False, overwrite: bool = False) -> typing.List[str]:
		"""
		Download a file or directory from the remote system using SFTP.

		:param path: Absolute path to the file or directory on the remote system.
		:param dst: The local destination file or directory to download to.
		:param directory_contents_only: If `True` and `path` points to a directory, then just the contents of the directory will be downloaded rather than the directory itself.
		:param overwrite: Whether to overwrite existing files.

		:return: A list of local paths to all files downloaded.
		"""
		assert self._connection and self._connection.client, "Connection is not open"
		log.info(f"{str(self)} Downloading \"{path}\" to \"{dst}\"...")
		assert os.path.isabs(path), f"{path} is not an absolute path"
		dst = os.path.abspath(dst)

		# Ensure the path exists
		path_stat = self.stat(path)
		if path_stat is None:
			raise exceptions.RemoteFileNotFoundError(self, path)

		if directory_contents_only and path_stat["isfile"]:
			raise exceptions.RemoteNotADirectoryError(self, path)

		# Get a mapping of remote paths to local paths
		# We need to test for different cases and handle each
		files_remote_to_local: typing.Dict[str, str] = {}
		directories_remote_to_local: typing.Dict[str, str] = {}

		if path_stat["isfile"]: # Single file download
			target = dst
			if not os.path.exists(target):
				# Target doesn't exist, we will download under the name specified
				# Just ensure the parent directory does exist
				parent = os.path.dirname(target)
				if not os.path.isdir(parent):
					raise NotADirectoryError(f"{parent} is not a valid local directory")
			else:
				# Target exists, we'll download into the target with the same name
				if not os.path.isdir(target):
					raise NotADirectoryError(f"{target} is not a valid local directory")
				target = os.path.join(target, os.path.basename(path))

			files_remote_to_local[path] = target

		elif path_stat["isdir"]: # Directory download
			target = dst
			if os.path.isdir(target) and directory_contents_only:
				# Downloading directory contents into a directory
				pass
			elif os.path.isdir(target) and not directory_contents_only:
				# Downloading directory into a directory, download with the same name into the target
				target = os.path.join(target, os.path.basename(path))
			elif not os.path.exists(target) and directory_contents_only:
				# We cannot download the contents of a directory if the dst does not exist
				raise NotADirectoryError(f"{dst} is not a valid local directory")
			elif not os.path.exists(target) and not directory_contents_only:
				# Downloading directory under a new name, we just need to ensure that the parent directory exists
				parent = os.path.dirname(target)
				if not os.path.isdir(parent):
					raise NotADirectoryError(f"{parent} is not a valid local directory")

			files_remote_to_local = { file: os.path.join(target, file[len(path):].strip("/")) for file in self.exec(f"find {path} -type f -print", assert_status=0).stdout.split("\n") }
			directories_remote_to_local = { directory: os.path.join(target, directory[len(path):].strip("/")) for directory in self.exec(f"find {path} -type d -print", assert_status=0).stdout.split("\n") }

		if len(files_remote_to_local) == 0 and len(directories_remote_to_local) == 0:
			return [] # Nothing to do

		# Ensure that no local copies exist (if overwrite is not True)
		if not overwrite:
			for local_path in files_remote_to_local.values():
				if os.path.exists(local_path):
					raise FileExistsError(f"{local_path} already exists")
				if os.path.isdir(local_path):
					raise IsADirectoryError(f"{local_path} cannot be overwritten as it exists as a directory")

		# Create directories
		for local_path in directories_remote_to_local.values():
			os.makedirs(local_path, exist_ok=True)

		# # Begin downloading files
		with self._connection.client.open_sftp() as ftp:
			for remote_path, local_path in files_remote_to_local.items():
				log.debug(f"{self} Downloading {remote_path} to {local_path}")
				ftp.get(remote_path, local_path)

		return list(files_remote_to_local.values())
			
	def upload(self, src: str, dst: str, directory_contents_only: bool = False, overwrite: bool = False) -> typing.List[str]:
		"""
		Upload a local file or directory to this path on the remote system using SFTP.

		:param src: The local path to a file or directory to upload.
		:param dst: The absolute remote path to upload to.
		:param directory_contents_only: If `True` and `src` points to a directory, then only the contents of the directory will be uploaded rather than the directory itself.
		:param overwrite: Whether to overwrite existing files on the remote system.

		:return: A list of remote paths to all files uploaded.
		"""
		assert self._connection and self._connection.client, "Connection is not open"
		log.info(f"<{self}> Uploading \"{src}\" to \"{dst}\"...")
		assert os.path.isabs(dst), f"{dst} is not an absolute path"
		src = os.path.abspath(src)
		if not os.path.exists(src):
			raise FileNotFoundError(f"File or directory {src} not found")
	
		if directory_contents_only and not os.path.isdir(src):
			raise NotADirectoryError(f"{src} is not a valid local directory")

		dst_stat = self.stat(dst)

		# Get a mapping of local paths to remote paths
		# We need to test for different cases and handle each
		files_local_to_remote: typing.Dict[str, str] = {}
		directories_local_to_remote: typing.Dict[str, str] = {}

		if os.path.isfile(src): # Single file upload
			target = dst
			if dst_stat is None:
				# Target doesn't exist, we will upload under the name specified
				# Just ensure the parent directory does exist
				parent = os.path.dirname(target)
				if not self.isdir(parent):
					raise exceptions.RemoteNotADirectoryError(self, parent)
			else:
				# Target exists, we'll upload into the target with the same name
				if dst_stat["isdir"]:
					target = os.path.join(target, os.path.basename(src))
				elif overwrite:
					parent = os.path.dirname(target)
					if not self.isdir(parent):
						raise exceptions.RemoteNotADirectoryError(self, parent)
				else:
					raise exceptions.RemoteFileExistsError(self, target)

			files_local_to_remote[src] = target.rstrip("/")

		elif os.path.isdir(src): # Directory upload
			target = dst
			target_exists = dst_stat is not None
			target_is_dir = dst_stat and dst_stat["isdir"]
			if target_is_dir and directory_contents_only:
				# Uploading directory contents into a directory
				pass
			elif target_is_dir and not directory_contents_only:
				# Uploading directory into a directory, upload with the same name into the target
				target = os.path.join(target, os.path.basename(src))
			elif not target_exists and directory_contents_only:
				# We cannot upload the contents of a directory if the dst does not exist
				raise exceptions.RemoteNotADirectoryError(self, target)
			elif not target_exists and not directory_contents_only:
				# Uploading directory under a new name, we just need to ensure that the parent directory exists
				parent = os.path.dirname(target)
				if not self.isdir(parent):
					raise exceptions.RemoteNotADirectoryError(self, parent)
			
			directories_local_to_remote[src] = target
			for root, directories, files in os.walk(src):
				for directory in directories:
					path = os.path.join(root, directory)
					directories_local_to_remote[path] = os.path.join(target, path[len(src):].lstrip("/")).rstrip("/")

				for file in files:
					path = os.path.join(root, file)
					files_local_to_remote[path] = os.path.join(target, path[len(src):].lstrip("/")).rstrip("/")

		if len(files_local_to_remote) == 0 and len(directories_local_to_remote) == 0:
			return [] # Nothing to do

		# Ensure that the upload will be valid (e.g. not overwriting files unless requested)
		# We only need to check in the case that the destination already exists
		if dst_stat is not None: 
			# Check the existence of all files and directories
			# Do this in a script so we can do this efficiently
			script = "\n".join([ f"stat -c '%F|%n' '{remote_path}' || true" for remote_path in [ *list(files_local_to_remote.values()), *list(directories_local_to_remote.values()) ] ])
			resp = self.exec_script(script, timeout=300)

			existing_remote_files = []
			existing_remote_directories = []
			for line in resp.stderr.strip().split("\n"):
				if not line:
					continue
				if "no such file" not in line.lower():
					raise exceptions.RemoteConnectionError(self, line)

			for line in resp.stdout.strip().split("\n"):
				if not line:
					continue
				filetype, filename = line.split("|", maxsplit=1)
				if "file" in filetype:
					existing_remote_files.append(filename.rstrip())
				elif "directory" in filetype:
					existing_remote_directories.append(filename.rstrip())

			for remote_file in files_local_to_remote.values():
				if remote_file in existing_remote_files and not overwrite:
					raise exceptions.RemoteFileExistsError(self, remote_file)
				if remote_file in existing_remote_directories:
					raise exceptions.RemoteIsADirectoryError(self, remote_file)

		# Determine the directories to create
		# We will be running mkdir with parents, so we only want the longest paths (to minimize the number of `mkdir` calls)
		dirs_to_make = []
		for remote_directory in directories_local_to_remote.values():
			dirname = remote_directory + "/"
			dirs_to_make = [ path for path in dirs_to_make if not dirname.startswith(path) ] # Remove any paths that are subsets of this path 
			is_parent = any([ path.startswith(dirname) for path in dirs_to_make ]) # Check if this is a parent of any existing path already
			if not is_parent:
				dirs_to_make.append(dirname)

		# Make directories
		# We will do this in a script to perform all operations in a single call
		# This is an optimization over making individual `mkdir` calls when the number of calls required is large
		script_contents = ""
		for dirname in dirs_to_make:
			script_contents += f"mkdir -p {dirname}\n"
		self.exec_script(script_contents, assert_status=0)
			
		# Begin uploading files
		with self._connection.client.open_sftp() as ftp:
			for local_path, remote_path in files_local_to_remote.items():
				log.debug(f"<{self}> Uploading \"{local_path}\" to \"{remote_path}\"")
				ftp.put(local_path, remote_path)

		return list(files_local_to_remote.values())

	def restart(self):
		"""
		Perform a graceful restart of the VM using `shutdown -r now`. This will close the current connection.
		"""
		self.exec("sudo shutdown -r now", pty=True)
		self.close()

	def shutdown(self):
		"""
		Attempt a graceful shutdown of the VM using `shutdown now`. This will close the current connection.
		"""
		self.exec("sudo shutdown now", pty=True)
		self.close()

	def time(self):
		"""
		Get the current time set on the remote host.

		:return: A `datetime` object.
		"""
		output = self.exec('date -u', assert_status=0).stdout.splitlines()[-1].strip()
		return parsetime(output).astimezone(datetime.timezone.utc)


from esxi_utils.util.response import Response, ResponsePromise
from esxi_utils.util import log, exceptions
from paramiko import ssh_exception
import paramiko
import fabric
import invoke
import random
import typing
import time
import re
import os 


class _ParamikoTransportWrapper:
	"""Transparent wrapper around a paramiko transport object."""

	# We define this object in order to allow us to set the pty width and height in fabric. Fabric does not offer a native method for setting
	# the pty width/height, and instead relies on the current terminal width/height (or a small default of 80/24). Since certain functionality
	# may rely on a minimum terminal size, we require a method to set these values.
	# To achieve this, we create a transparent wrapper object around fabric's underlying paramiko transport object. For most functions,
	# this wrapper will do nothing, but for the `open_session` function we'll instead return another wrapper around the channel object.
	# By doing this, we can ultimately overwrite the channel's `get_pty` function to enforce the minimum pty width/height.
	# This monkey-patch is the simplest and most forward-compatible method to achieve this functionality without rewriting significant portions 
	# of fabric's code, or switching libraries entirely.
	# Unfortunately, setting values for this method suffers from not being thread-safe, but due to the design of fabric's ``run`` method there appears to be
	# no way to create a thread-safe solution without a significant rewrite of both fabric and invoke. Thus, these values should not be set dynamically
	# (i.e. per-command), and instead should only be set once for the entire connection lifetime.

	def __init__(self, obj: 'paramiko.Transport', min_pty_width: int, min_pty_height: int):
		self._wrapped_obj = obj
		self._min_pty_width = min_pty_width
		self._min_pty_height = min_pty_height

	def __getattr__(self, attr):
		if attr in self.__dict__:
			return getattr(self, attr)
		return getattr(self._wrapped_obj, attr)
	
	def open_session(self, *args, **kwargs):
		return _ParamikoChannelWrapper(self._wrapped_obj.open_session(*args, **kwargs), self._min_pty_width, self._min_pty_height)

	def open_channel(self, *args, **kwargs):
		return _ParamikoChannelWrapper(self._wrapped_obj.open_channel(*args, **kwargs), self._min_pty_width, self._min_pty_height)


class _ParamikoChannelWrapper:
	"""Transparent wrapper around a paramiko channel object."""
	def __init__(self, obj: 'paramiko.Channel', min_pty_width: int, min_pty_height: int):
		self._wrapped_obj = obj
		self._min_pty_width = min_pty_width
		self._min_pty_height = min_pty_height

	def __getattr__(self, attr):
		if attr in self.__dict__:
			return getattr(self, attr)
		return getattr(self._wrapped_obj, attr)
	
	def get_pty(self, term='vt100', width=80, height=24, width_pixels=0, height_pixels=0):
		width = max(self._min_pty_width, width)
		height = max(self._min_pty_height, height)
		self._wrapped_obj.get_pty(term=term, width=width, height=height, width_pixels=width_pixels, height_pixels=height_pixels)


class SSHConnection:
	"""
	A basic SSH connection to a remote host.

	:param ip: The IP of the target.
	:param username: The username to use when logging into the remote system.
	:param password: The password to use when logging into the remote system.
	:param port: The port to use when connecting over SSH.
	:param min_pty_width: The minimum width of pseudoterminals (in columns) when ``pty=True`` is set on a command.
	:param min_pty_height: The minimum height of pseudoterminals (in rows) when ``pty=True`` is set on a command.
	"""
	def __init__(self, ip: str, username: str, password: str, port: int = 22, min_pty_width: int = 160, min_pty_height: int = 48):
		os.environ['SSH_AUTH_SOCK'] = ''
		self._ip = ip
		self._username = username
		self._password = password
		self._port = port
		self._connection: typing.Optional[fabric.Connection] = None
		self._min_pty_width = min_pty_width
		self._min_pty_height = min_pty_height

	def __enter__(self):
		self.open()
		return self

	def __exit__(self, *args):
		self.close()

	def open(self):
		"""
		Open a new connection. If an connection is already open, it will be closed before opening a new one.
		"""
		if self._connection:
			self.close()
		self._connection = fabric.Connection(self._ip, user=self._username, port=self._port, connect_kwargs={ "password": self._password, "timeout": 60, "banner_timeout": 60, "auth_timeout": 60, "look_for_keys": False }, inline_ssh_env=True)
		assert self._connection.client, "Unable to get connection client"
		self._connection.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

		err = None
		for _ in range(3):
			try:
				self._connection.open()
				assert self._connection.transport, "Unable to get connection transport"
				self._connection.transport.set_keepalive(30)

				# Wrap the transport in our wrapper object
				self._connection.transport = _ParamikoTransportWrapper(self._connection.transport, self._min_pty_width, self._min_pty_height) # type: ignore
				return
			except ssh_exception.AuthenticationException as auth_err:
				self.close()
				raise auth_err
			except Exception as e:
				# Opening connections may rarely fail for various reasons, especially when the network is congested
				# Waiting a random period of time and trying again typically resolves this
				err = e
				time.sleep(random.randint(2, 5))
		self.close()
		raise exceptions.RemoteConnectionError(self, err)

	def close(self):
		"""
		Terminate the network connection to the remote end, if open.

		If no connection is open, this method does nothing.
		"""
		if self._connection:
			self._connection.close()
			self._connection = None

	def wait(self, retries: int = 60, delay: int = 2, keep_open=False) -> bool:
		"""
		Waits until this connection can be established, and then establishes the connection.

		:param retries: How many times to retry connecting before exiting. Authentication errors will terminate the wait early.
		:param delay: How long to pause between retries in seconds.
		:param keep_open: Whether or not the keep the connection open. If `True`, it is left to the user to close the connection.

		:return: Whether or not the connection could be established
		"""
		fault_tolerance = 1 # When a VM is booting, we may get false errors that resolve with time; allow this many hard failures before exiting
		retries = retries + 1
		for i in range(retries):
			if i != 0:
				time.sleep(delay)
			try:
				self.open()
				if not keep_open:
					self.close()
				return True
			except (ssh_exception.BadAuthenticationType, ssh_exception.PartialAuthentication):
				continue
			except ssh_exception.AuthenticationException as e:
				if fault_tolerance == 0:
					raise e
				else:
					fault_tolerance -= 1
					time.sleep(random.randint(20, 40))
			except Exception:
				continue
		return False

	def exec_async(
		self, 
		cmd: str, 
		stdin: typing.Optional[typing.Dict[str, str]] = None, 
		timeout: int = 120, 
		pty: bool = False,
		cwd: typing.Optional[str] = None, 
		env: typing.Optional[typing.Dict[str, typing.Any]] = None, 
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None,
	) -> 'SSHResponsePromise':
		"""
		Execute a command asynchronously on the remote end of this connection.

		:param cmd: The command to run.
		:param stdin: A dict mapping a 'pattern' (regex string) to a 'response' (string), to input 'response' to stdin when stdout matches 'pattern'
		:param timeout: The command timeout in seconds. Set to 0 to disable timeout.
		:param pty: Run the command in a pseudoterminal. If `True`, the returned `Response` object will have a merged stdout and stderr, and stderr will be empty.
		:param cwd: The directory that the command should run in. If `None`, runs in the default directory.
		:param env: A dict of shell environment variables to be merged into the default environment that the remote command executes within.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 

		:return: A `SSHResponsePromise` object.
		"""
		if self._connection is None:
			raise exceptions.RemoteConnectionNotOpenError(self)
		assert isinstance(cmd, str) and len(cmd.strip()) != 0, "cmd must be a non-empty string"
		assert isinstance(stdin, dict) or stdin is None, "stdin must be a dict or None"
		assert isinstance(timeout, int) and timeout >= 0, "timeout must be an integer greater or equal to 0"
		assert isinstance(pty, bool), "pty must be a boolean"
		assert callable(out_stream_callback) or out_stream_callback is None, "out_stream_callback must be a function or None"
		
		stdin = stdin or dict()
		env = env or dict()

		# Create watchers
		watchers = [ invoke.Responder(pattern=pattern, response=response) for pattern, response in stdin.items() ]

		# If out_stream_callback specified, create the handler
		out_stream = None
		if out_stream_callback:
			class OutStreamHandler:
				def flush(self):
					pass
				def write(self, s):
					if out_stream_callback:
						out_stream_callback(s)
					return len(s)
			out_stream = OutStreamHandler()
		
		log.info(f"{str(self)} Running SSH command: {cmd}")
		kwargs = { "hide": True, "warn": True, "pty": pty, "env": env, "replace_env": True, "timeout": timeout, "watchers": watchers, "out_stream": out_stream, "asynchronous": True }
		
		if cwd:
			with self._connection.cd(cwd):
				promise: invoke.runners.Promise = self._connection.run(cmd, **kwargs)
				return SSHResponsePromise(self, promise)
			
		promise: invoke.runners.Promise = self._connection.run(cmd, **kwargs)
		return SSHResponsePromise(self, promise)

	def exec(
		self, 
		cmd: str, 
		stdin: typing.Optional[typing.Dict[str, str]] = None, 
		timeout: int = 120, 
		pty: bool = False,
		cwd: typing.Optional[str] = None, 
		env: typing.Optional[typing.Dict[str, typing.Any]] = None, 
		out_stream_callback: typing.Optional[typing.Callable[[str], None]] = None,
		assert_status: typing.Optional[int] = None,
		network_retries: int = 0,
	) -> 'Response':
		"""
		Execute a command on the remote end of this connection. Unlike ``exec_async``, this will block until the associated command has finished.

		:param cmd: The command to run.
		:param stdin: A dict mapping a 'pattern' (regex string) to a 'response' (string), to input 'response' to stdin when stdout matches 'pattern'
		:param timeout: The command timeout in seconds. Set to 0 to disable timeout.
		:param pty: Run the command in a pseudoterminal. If `True`, the returned `Response` object will have a merged stdout and stderr, and stderr will be empty.
		:param cwd: The directory that the command should run in. If `None`, runs in the default directory.
		:param env: A dict of shell environment variables to be merged into the default environment that the remote command executes within.
		:param out_stream_callback: Optional callback to handle the stdout of commands. If not `None`, this will be called as a command's stdout becomes available. 
		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.
		:param network_retries: Number of times to retry the command when a recoverable network error is encountered. The command should be idempotent when this is non-zero as it may run several times.

		:return: A `Response` object.
		"""
		run = lambda: self.exec_async(cmd=cmd, stdin=stdin, timeout=timeout, pty=pty, cwd=cwd, env=env, out_stream_callback=out_stream_callback).wait(assert_status)
		for _ in range(network_retries):
			try:
				return run()
			except EOFError:
				pass
		return run()
	
	def __str__(self):
		return f"<{type(self).__name__} {self._username}@{self._ip}>"

	def __repr__(self):
		return str(self)


class SSHResponsePromise(ResponsePromise):
	"""
	Encapsulates a Promise for a result of an SSH command.

	:param connection: The SSHConnection instance that generated this promise.
	:param promise: The Promise object for the command.
	"""
	def __init__(self, connection: SSHConnection, promise: invoke.runners.Promise):
		super().__init__()
		self._connection = connection
		self._promise = promise
		self._cmd = self._promise.command
		log.debug(f"New SSH Promise Object: {str(self)}")

	def wait(self, assert_status: typing.Optional[int] = None) -> Response:
		"""
		Block until the associated command exits, returning the result. 

		:param assert_status: Assert that the command exits with a certain status number. If `None`, no assertion is made and checking the status is left to the caller.
		"""
		super().wait()
		assert isinstance(assert_status, int) or assert_status is None, "assert_status must be an integer or None"
		
		resp = self._promise.join()

		stdout = resp.stdout.strip()
		stderr = resp.stderr.strip()

		# If run as sudo, remove the sudo prompt from output
		if self._cmd.startswith("sudo"):
			stdout = re.sub(r"^\[sudo\] password[^\n]+\n?", "", stdout)

		result = Response(cmd=self._cmd, stdout=stdout, stderr=stderr, status=resp.exited)
		if assert_status is not None and assert_status != result.status:
			raise exceptions.RemoteConnectionCommandError(self._connection, result)
		return result

	def cancel(self):
		"""
		Cancel the underlying operation.
		"""
		super().cancel()
		self._promise.runner.stop()
		try:
			self._promise.join()
		except invoke.exceptions.CommandTimedOut:
			pass

	def __str__(self):
		return repr(self)

	def __repr__(self):
		return f"<{type(self).__name__} for \"{self._cmd}\">"


from esxi_utils.vm.tools.guesttools import GuestTools
from esxi_utils.util.response import Response
from esxi_utils.util import exceptions
from esxi_utils.util import log
import pyVmomi
import time
import typing


class PanosGuestTools(GuestTools):
	"""
	Guest tools subclass specifically for interaction with Palo Alto systems.

	Wrapper class for functionality related to VMware Tools on a Virtual Machine. The virtual machine must have VMware tools installed for
	any of the contained functions to work.

	:param vm: A `VirtualMachine` object to wrap. VMware tools must be installed on this virtual machine.
	"""
	def execute_panos_cmd(
		self,
		username: str,
		password: str,
		cmd_list: list,
		list_delimiter: str = '%s\n',
		output_pipe_cmd: str = '',
		bash_path: str = "/bin/bash",
		timeout: typing.Optional[int] = 180,
		retry_timeout_minutes: int = 10
	) -> 'Response':
		"""
		Emulates sending commands on the Palo Alto CLI by piping a list of command strings into the Palo Alto CLI binary.
		Bash will be used as the underlying shell for interaction with the Palo Alto CLI binary.
		Systems without bash will fail. Sudo is not supported on Palo Alto machines.

		:param username:
			The username of the virtual machine user.
		:param password:
			The password of the virtual machine user.
		:param cmd_list:
			A list of 'Palo Alto CLI' commands to execute. This function will raise an exception if an empty list is given.
		:param list_delimiter:
			How to separate commands in the list. Default is 'newline' separation, which simulates pressing enter on the CLI itself.
		:param output_pipe_cmd:
			An optional command (or series of commands) to pipe the output of the CLI into. THESE ARE BASH COMMANDS. This will not be delimited.
		:param bash_path:
			The path to the bash executable.
		:param timeout:
			The command timeout in seconds. Set to 0 or `None` to disable timeout.
		:param retry_timeout_minutes:
			Certain exceptions can cause a 'retry' of the command:
			This value represents how many minutes to keep 'retrying' for.
		:return:
			A `types.Response` object
		"""
		self._assert_available()
		# Throw an error if the list is empty (the user likely expects a response of some kind, error prone to return anything)
		if len(cmd_list) <= 0:
			raise exceptions.GuestToolsError(self._vm, "ERROR: 'execute_panos_cmd' was given an empty list of commands to execute!")
		# Use printf with a given parameter delimiter/separator (typically newline) to pipe commands into the CLI binary
		cmd = f'printf \'{list_delimiter}\''
		if not isinstance(cmd_list, list):
			cmd_list = [cmd_list]
		# Build the command
		for entry in cmd_list:
			cmd += " '" + entry + "' "
		# Build the pipe into the binary
		cmd += "| /usr/local/bin/cli"
		# Pipe out into any additional commands
		if output_pipe_cmd != '':
			cmd += ' | ' + output_pipe_cmd

		timeout_time = retry_timeout_minutes * 60
		start_time = time.time()
		while True:
			try:
				with self.use_temporary_file(username=username, password=password, extension=".sh") as script_path:
					self.write_file(username=username, password=password, filepath=script_path, data=cmd)
					response = self.execute_program(
						username=username,
						password=password,
						command=f"{bash_path} {script_path}",
						timeout=timeout
					)
					return Response(cmd=cmd, stdout=response.stdout, stderr=response.stderr, status=response.status)
			except pyVmomi.vmodl.fault.SystemError as e:
				log.debug(f'\nWARN: Caught pyVmomi "vmodl.fault.SystemError" exception while trying to execute_panos_cmd! Will ignore and try again. Error: {str(e)} ... Retrying...')
			except pyVmomi.vim.fault.InvalidState as e2:
				log.debug(f'\nWARN: Caught pyVmomi "vim.fault.InvalidState" exception while trying to execute_panos_cmd! Will ignore and try again. Error: {str(e2)} ... Retrying...')
			if start_time + timeout_time <= time.time():
				raise Exception('ERROR: Timeout while trying to execute panos command! The "cmd" variable: ' + cmd)
			# end try/except
		# end while loop
	# end fn execute_panos_cmd

	def show_system_info(self, username: str, password: str) -> 'Response':
		"""
		Executes the command 'show system info' on the Palo Alto CLI.

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.

		:return: A `types.Response` object
		"""
		return self.execute_panos_cmd(username, password, ['show system info'])

	def get_ip_address(self, username: str, password: str) -> 'Response':
		"""
		Executes the command 'show system info' on the Palo Alto CLI and then does a linux 'grep' for 'ip-address'

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.

		:return: A `types.Response` object
		"""
		return self.execute_panos_cmd(username, password, ['show system info'], output_pipe_cmd='grep ip-address')

	def set_ip_address(
		self,
		username: str,
		password: str,
		ip_address: str,
		netmask: str,
		default_gateway: typing.Optional[str],
		type: str = 'static'
	) -> 'Response':
		"""
		Executes a series of Palo Alto CLI commands to change the IP address of the VM.
		- enters 'configure' mode
		- executes 'set deviceconfig system ip-address....' and provides the given values as arguments
		- commits changes

		:param username:
			The username of the virtual machine user.
		:param password:
			The password of the virtual machine user.
		:param ip_address:
			The IP address to assign to the VM.
		:param netmask:
			The netmask to apply to the VM.
		:param default_gateway:
			The default gateway to apply the VM (can be None).
		:param type:
			Known good value is 'static' (as in 'static IP'). Change at your own risk.
		:return:
			A `types.Response` object
		"""
		gateway_str = f' default-gateway {default_gateway}'
		if default_gateway is None:
			gateway_str = ''
		ip_config_cmd = f'set deviceconfig system ip-address {ip_address} netmask {netmask}{gateway_str} type {type}'
		return self.execute_panos_cmd(username, password, ['configure t', ip_config_cmd, 'commit force'])

	def set_password(self, username: str, old_password: str, new_password: str) -> 'Response':
		"""
		Executes a series of Palo Alto CLI commands to chage the password for the given username.
		- enters 'configure' mode
		- executes 'set mgt-config users admin password'
		- inputs the new_password (twice)
		- commits changes

		:param username: The username of the virtual machine user. This username's password will be changed.
		:param old_password: The password of the virtual machine user.
		:param new_password: The password to give to the provided username.

		:return: A `types.Response` object
		"""
		return self.execute_panos_cmd(username, old_password, ['configure t', f'set mgt-config users {username} password', new_password, new_password, 'commit force'])


	def create_new_user(
		self,
		existing_admin_username: str,
		existing_admin_password: str,
		new_username: str,
		new_password: str
	) -> 'Response':
		"""
		Executes a series of Palo Alto CLI commands to create a new user on the VM.
		- enters 'configure' mode
		- executes 'set mgt-config users <username> password'
		- inputs the new_password (twice)
		- commits changes

		:param existing_admin_username:
			The username of the virtual machine user admin that already exists.
		:param existing_admin_password:
			The password of the virtual machine user admin that already exists.
		:param new_username:
			The username of the virtual machine user to create.
		:param new_password:
			The password to give to the provided username.

		:return:
			A `types.Response` object for the executed command
		"""
		return self.execute_panos_cmd(existing_admin_username, existing_admin_password, ['configure t', f'set mgt-config users {new_username} password', new_password, new_password, 'commit force']) 


	def give_user_superuser_rights(
		self,
		existing_admin_username: str,
		existing_admin_password: str,
		username_to_give_rights_to: str,
		readonly: bool = False
	) -> 'Response':
		"""
		Executes a series of Palo Alto CLI commands to give superuser rights to the provided username.
		- enters 'configure' mode
		- executes 'set mgt-config users <username> permissions role-based super<user/reader> yes'
		- commits changes

		:param existing_admin_username:
			The username of the virtual machine user admin that already exists.
		:param existing_admin_password:
			The password of the virtual machine user admin that already exists.
		:param username_to_give_rights_to:
			The username of the virtual machine user to give rights to
		:param readonly:
			Assign the role 'superreader' instead of 'superuser'

		:return: A `types.Response` object for the executed command
		"""
		if readonly:
			role = 'superreader'
		else:
			role = 'superuser'
		return self.execute_panos_cmd(existing_admin_username, existing_admin_password, ['configure t', f'set mgt-config users {username_to_give_rights_to} permissions role-based {role} yes', 'commit force'])


	def enable_server_verification(self, username: str, password: str, enable: bool = True, timeout: int = 180) -> 'Response':
		"""
		Executes a series of Palo Alto CLI commands to enable server-verification.
		- enters 'configure' mode
		- executes 'set deviceconfig system server-verification yes'
		- commits changes

		:param username: The username of the virtual machine user.
		:param password: The password of the virtual machine user.
		:param enable: When set to True: enables server-verification. When False: disables server-verification (not recommended)
		:param timeout: How long in seconds to wait for retry.
		
		:return: A `types.Response` object
		"""
		cmd = 'set deviceconfig system server-verification '
		if enable:
			cmd += 'yes'
		else:
			cmd += 'no'
		return self.execute_panos_cmd(username, password, ['configure', cmd, 'commit force'], timeout=timeout)


	def load_configuration_file(self, non_default_username: str, password: str, filename: str) -> 'Response':
		"""
		Executes a series of Palo Alto CLI commands to load an XML configuration file.
		- enters 'configure' mode
		- executes 'load config from <filename>'
		- commits changes

		:param non_default_username:
			The username of a VM 'superuser' that isn't the default user (typically 'admin').
			This is important because the default username typically has the password overwritten.
		:param password:
			The password for the 'non_default_username'.
		:param filename:
			The name of the file (NOT the path) to load.

		:return: A `types.Response` object for the executed command
		"""
		return self.execute_panos_cmd(non_default_username, password, ['configure t', f'load config from {filename}', 'commit force'])
